""" Callbacks related to training progress. """

import os
import pickle
import warnings
from contextlib import suppress
from fnmatch import fnmatch
from functools import partial
from itertools import product
from copy import deepcopy

import numpy as np
from skorch.callbacks import Callback
from skorch.exceptions import SkorchException
from skorch.utils import _check_f_arguments
from skorch.utils import noop
from skorch.utils import open_file_like
from skorch.utils import freeze_parameter
from skorch.utils import unfreeze_parameter


__all__ = ['Checkpoint', 'EarlyStopping', 'ParamMapper', 'Freezer',
           'Unfreezer', 'Initializer', 'InputShapeSetter', 'LoadInitState',
           'TrainEndCheckpoint']


class Checkpoint(Callback):
    """Save the model during training if the given metric improved.

    This callback works by default in conjunction with the validation
    scoring callback since it creates a ``valid_loss_best`` value
    in the history which the callback uses to determine if this
    epoch is save-worthy.

    You can also specify your own metric to monitor or supply a
    callback that dynamically evaluates whether the model should
    be saved in this epoch.

    As checkpointing is often used in conjunction with early stopping
    there is a need to restore the state of the model to the best
    checkpoint after training is done. The checkpoint callback will
    do this for you if you wish.

    Some or all of the following can be saved:

      - model parameters (see ``f_params`` parameter);
      - optimizer state (see ``f_optimizer`` parameter);
      - criterion state (see ``f_criterion`` parameter);
      - training history (see ``f_history`` parameter);
      - entire model object (see ``f_pickle`` parameter).

    If you've created a custom module, e.g. ``net.mymodule_``, you
    can save that as well by passing ``f_mymodule``.

    You can implement your own save protocol by subclassing
    ``Checkpoint`` and overriding :func:`~Checkpoint.save_model`.

    This callback writes a bool flag to the history column
    ``event_cp`` indicating whether a checkpoint was created or not.

    Example:

    >>> net = MyNet(callbacks=[Checkpoint()])
    >>> net.fit(X, y)

    Example using a custom monitor where models are saved only in
    epochs where the validation *and* the train losses are best:

    >>> monitor = lambda net: all(net.history[-1, (
    ...     'train_loss_best', 'valid_loss_best')])
    >>> net = MyNet(callbacks=[Checkpoint(monitor=monitor)])
    >>> net.fit(X, y)

    Parameters
    ----------
    monitor : str, function, None
      Value of the history to monitor or callback that determines
      whether this epoch should lead to a checkpoint. The callback
      takes the network instance as parameter.

      In case ``monitor`` is set to ``None``, the callback will save
      the network at every epoch.

      **Note:** If you supply a lambda expression as monitor, you cannot
      pickle the wrapper anymore as lambdas cannot be pickled. You can
      mitigate this problem by using importable functions instead.

    f_params : file-like object, str, None (default='params.pt')
      File path to the file or file-like object where the model
      parameters should be saved. Pass ``None`` to disable saving
      model parameters.

      If the value is a string you can also use format specifiers
      to, for example, indicate the current epoch. Accessible format
      values are ``net``, ``last_epoch`` and ``last_batch``.
      Example to include last epoch number in file name:

      >>> cb = Checkpoint(f_params="params_{last_epoch[epoch]}.pt")

    f_optimizer : file-like object, str, None (default='optimizer.pt')
      File path to the file or file-like object where the optimizer
      state should be saved. Pass ``None`` to disable saving
      model parameters.

      Supports the same format specifiers as ``f_params``.

    f_criterion : file-like object, str, None (default='criterion.pt')
      File path to the file or file-like object where the criterion
      state should be saved. Pass ``None`` to disable saving
      model parameters.

      Supports the same format specifiers as ``f_params``.

    f_history : file-like object, str, None (default='history.json')
      File path to the file or file-like object where the model
      training history should be saved. Pass ``None`` to disable
      saving history.

    f_pickle : file-like object, str, None (default=None)
      File path to the file or file-like object where the entire
      model object should be pickled. Pass ``None`` to disable
      pickling.

      Supports the same format specifiers as ``f_params``.

    fn_prefix: str (default='')
      Prefix for filenames. If ``f_params``, ``f_optimizer``, ``f_history``,
      or ``f_pickle`` are strings, they will be prefixed by ``fn_prefix``.

    dirname: str (default='')
      Directory where files are stored.

    load_best: bool (default=False)
      Load the best checkpoint automatically once training ended.
      This can be particularly helpful in combination with early stopping
      as it allows for scoring with the best model, even when early stopping
      ended training a number of epochs later. Note that this will only
      work when ``monitor != None``.

    event_name: str, (default='event_cp')
      Name of event to be placed in history when checkpoint is triggered.
      Pass ``None`` to disable placing events in history.

    sink : callable (default=noop)
      The target that the information about created checkpoints is
      sent to. This can be a logger or ``print`` function (to send to
      stdout). By default the output is discarded.

    """
    def __init__(
            self,
            monitor='valid_loss_best',
            f_params='params.pt',
            f_optimizer='optimizer.pt',
            f_criterion='criterion.pt',
            f_history='history.json',
            f_pickle=None,
            fn_prefix='',
            dirname='',
            event_name='event_cp',
            sink=noop,
            load_best=False,
            **kwargs
    ):
        self.monitor = monitor
        self.f_params = f_params
        self.f_optimizer = f_optimizer
        self.f_criterion = f_criterion
        self.f_history = f_history
        self.f_pickle = f_pickle
        self.fn_prefix = fn_prefix
        self.dirname = dirname
        self.event_name = event_name
        self.sink = sink
        self.load_best = load_best
        self._check_kwargs(kwargs)
        vars(self).update(**kwargs)
        self._validate_filenames()

    def _check_kwargs(self, kwargs):
        for key in kwargs:
            if not key.startswith('f_'):
                raise TypeError(
                    "{cls_name} got an unexpected argument '{key}', did you mean "
                    "'f_{key}'?".format(cls_name=self.__class__.__name__, key=key))

    def initialize(self):
        self._validate_filenames()
        if self.dirname and not os.path.exists(self.dirname):
            os.makedirs(self.dirname, exist_ok=True)
        return self

    def on_train_end(self, net, **kwargs):
        if not self.load_best or self.monitor is None:
            return
        self._sink("Loading best checkpoint after training.", net.verbose)
        net.load_params(checkpoint=self)

    def on_epoch_end(self, net, **kwargs):
        if "{}_best".format(self.monitor) in net.history[-1]:
            warnings.warn(
                "Checkpoint monitor parameter is set to '{0}' and the history "
                "contains '{0}_best'. Perhaps you meant to set the parameter "
                "to '{0}_best'".format(self.monitor), UserWarning)

        if self.monitor is None:
            do_checkpoint = True
        elif callable(self.monitor):
            do_checkpoint = self.monitor(net)
        else:
            try:
                do_checkpoint = net.history[-1, self.monitor]
            except KeyError as e:
                msg = (
                    f"{e.args[0]} Make sure you have validation data if you use "
                    "validation scores for checkpointing.")
                raise SkorchException(msg)

        if self.event_name is not None:
            net.history.record(self.event_name, bool(do_checkpoint))

        if do_checkpoint:
            self.save_model(net)
            self._sink("A checkpoint was triggered in epoch {}.".format(
                len(net.history) + 1
            ), net.verbose)

    def _f_kwargs(self):
        return {key: getattr(self, key) for key in dir(self)
                if key.startswith('f_') and (key != 'f_history_')}

    def save_model(self, net):
        """Save the model.

        This function saves some or all of the following:

          - model parameters;
          - optimizer state;
          - criterion state;
          - training history;
          - custom modules;
          - entire model object.

        """
        kwargs_module, kwargs_other = _check_f_arguments(
            self.__class__.__name__, **self._f_kwargs())

        for key, val in kwargs_module.items():
            if val is None:
                continue

            f = self._format_target(net, val, -1)
            key = key[:-1]  # remove trailing '_'
            self._save_params(f, net, 'f_' + key, key + " state")

        f_history = kwargs_other.get('f_history')
        if f_history is not None:
            f = self.f_history_
            self._save_params(f, net, "f_history", "history")

        f_pickle = kwargs_other.get('f_pickle')
        if f_pickle:
            f_pickle = self._format_target(net, f_pickle, -1)
            with open_file_like(f_pickle, 'wb') as f:
                pickle.dump(net, f)

    @property
    def f_history_(self):
        # This is a property and not in initialize to allow ``NeuralNet``
        # to call ``load_params`` without needing the checkpoint to
        # by initialized.
        if self.f_history is None:
            return None
        return os.path.join(
            self.dirname, self.fn_prefix + self.f_history)

    def get_formatted_files(self, net):
        """Returns a dictionary of formatted filenames"""
        idx = -1
        if (
                self.event_name is not None and
                net.history
        ):
            for i, v in enumerate(net.history[:, self.event_name]):
                if v:
                    idx = i

        return {key: self._format_target(net, val, idx) for key, val
                in self._f_kwargs().items()}

    def _save_params(self, f, net, f_name, log_name):
        try:
            net.save_params(**{f_name: f})
        except Exception as e:  # pylint: disable=broad-except
            self._sink(
                "Unable to save {} to {}, {}: {}".format(
                    log_name, f, type(e).__name__, e), net.verbose)

    def _format_target(self, net, f, idx):
        """Apply formatting to the target filename template."""
        if f is None:
            return None
        if isinstance(f, str):
            f = self.fn_prefix + f.format(
                net=net,
                last_epoch=net.history[idx],
                last_batch=net.history[idx, 'batches', -1],
            )
            return os.path.join(self.dirname, f)
        return f

    def _validate_filenames(self):
        """Checks if passed filenames are valid.

        Specifically, f_* parameter should not be passed in
        conjunction with dirname.

        """
        _check_f_arguments(self.__class__.__name__, **self._f_kwargs())

        if not self.dirname:
            return

        def _is_truthy_and_not_str(f):
            return f and not isinstance(f, str)

        if any(_is_truthy_and_not_str(val) for val in self._f_kwargs().values()):
            raise SkorchException(
                'dirname can only be used when f_* are strings')

    def _sink(self, text, verbose):
        #  We do not want to be affected by verbosity if sink is not print
        if (self.sink is not print) or verbose:
            self.sink(text)


class EarlyStopping(Callback):
    """Callback for stopping training when scores don't improve.

    Stop training early if a specified `monitor` metric did not
    improve in `patience` number of epochs by at least `threshold`.

    Parameters
    ----------
    monitor : str (default='valid_loss')
      Value of the history to monitor to decide whether to stop
      training or not.  The value is expected to be double and is
      commonly provided by scoring callbacks such as
      :class:`skorch.callbacks.EpochScoring`.

    lower_is_better : bool (default=True)
      Whether lower scores should be considered better or worse.

    patience : int (default=5)
      Number of epochs to wait for improvement of the monitor value
      until the training process is stopped.

    threshold : int (default=1e-4)
      Ignore score improvements smaller than `threshold`.

    threshold_mode : str (default='rel')
        One of `rel`, `abs`. Decides whether the `threshold` value is
        interpreted in absolute terms or as a fraction of the best
        score so far (relative)

    sink : callable (default=print)
      The target that the information about early stopping is
      sent to. By default, the output is printed to stdout, but the
      sink could also be a logger or :func:`~skorch.utils.noop`.

    load_best: bool (default=False)
      Whether to restore module weights from the epoch with the best value of
      the monitored quantity. If False, the module weights obtained at the
      last step of training are used. Note that only the module is restored.
      Use the ``Checkpoint`` callback with the :attr:`~Checkpoint.load_best`
      argument set to ``True`` if you need to restore the whole object.

    """
    def __init__(
            self,
            monitor='valid_loss',
            patience=5,
            threshold=1e-4,
            threshold_mode='rel',
            lower_is_better=True,
            sink=print,
            load_best=False,
    ):
        self.monitor = monitor
        self.lower_is_better = lower_is_better
        self.patience = patience
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.misses_ = 0
        self.dynamic_threshold_ = None
        self.sink = sink
        self.load_best = load_best

    def __getstate__(self):
        # Avoids to save the module_ weights twice when pickling
        state = self.__dict__.copy()
        state['best_model_weights_'] = None
        return state

    # pylint: disable=arguments-differ
    def on_train_begin(self, net, **kwargs):
        if self.threshold_mode not in ['rel', 'abs']:
            raise ValueError("Invalid threshold mode: '{}'"
                             .format(self.threshold_mode))
        self.misses_ = 0
        self.dynamic_threshold_ = np.inf if self.lower_is_better else -np.inf
        self.best_model_weights_ = None
        self.best_epoch_ = 0

    def on_epoch_end(self, net, **kwargs):
        current_score = net.history[-1, self.monitor]
        if not self._is_score_improved(current_score):
            self.misses_ += 1
        else:
            self.misses_ = 0
            self.dynamic_threshold_ = self._calc_new_threshold(current_score)
            self.best_epoch_ = net.history[-1, "epoch"]
            if self.load_best:
                self.best_model_weights_ = deepcopy(net.module_.state_dict())
        if self.misses_ == self.patience:
            if net.verbose:
                self._sink("Stopping since {} has not improved in the last "
                           "{} epochs.".format(self.monitor, self.patience),
                           verbose=net.verbose)
            raise KeyboardInterrupt

    def on_train_end(self, net, **kwargs):
        if (
            self.load_best and (self.best_epoch_ != net.history[-1, "epoch"])
            and (self.best_model_weights_ is not None)
        ):
            net.module_.load_state_dict(self.best_model_weights_)
            self._sink("Restoring best model from epoch {}.".format(
                self.best_epoch_
            ), verbose=net.verbose)

    def _is_score_improved(self, score):
        if self.lower_is_better:
            return score < self.dynamic_threshold_
        return score > self.dynamic_threshold_

    def _calc_new_threshold(self, score):
        """Determine threshold based on score."""
        if self.threshold_mode == 'rel':
            abs_threshold_change = self.threshold * score
        else:
            abs_threshold_change = self.threshold

        if self.lower_is_better:
            new_threshold = score - abs_threshold_change
        else:
            new_threshold = score + abs_threshold_change
        return new_threshold

    def _sink(self, text, verbose):
        #  We do not want to be affected by verbosity if sink is not print
        if (self.sink is not print) or verbose:
            self.sink(text)


class ParamMapper(Callback):
    """Map arbitrary functions over module parameters filtered by pattern
    matching.

    In the simplest case the function is only applied once at
    the beginning of a given epoch (at ``on_epoch_begin``) but more complex
    execution schemes (e.g. periodic application) are possible using
    ``at`` and ``scheduler``.

    Notes
    -----
    When starting the training process after saving and loading a model,
    ``ParamMapper`` might re-initialize parts of your model when the
    history is not saved along with the model. To avoid this, in case
    you use ``ParamMapper`` (or subclasses, e.g. :class:`.Initializer`)
    and want to save your model make sure to either (a) use pickle,
    (b) save and load the history or (c) remove the parameter mapper
    callbacks before continuing training.

    Examples
    --------
    Initialize a layer on first epoch before the first training step:

    >>> init = partial(torch.nn.init.uniform_, a=0, b=1)
    >>> cb = ParamMapper('linear*.weight', at=1, fn=init)
    >>> net = Net(myModule, callbacks=[cb])

    Reset layer initialization if train loss reaches a certain value
    (e.g. re-initialize on overfit):

    >>> at = lambda net: net.history[-1, 'train_loss'] < 0.1
    >>> init = partial(torch.nn.init.uniform_, a=0, b=1)
    >>> cb = ParamMapper('linear0.weight', at=at, fn=init)
    >>> net = Net(myModule, callbacks=[cb])

    Periodically freeze and unfreeze all embedding layers:

    >>> def my_sched(net):
    ...    if len(net.history) % 2 == 0:
    ...        return skorch.utils.freeze_parameter
    ...    else:
    ...        return skorch.utils.unfreeze_parameter
    >>> cb = ParamMapper('embedding*.weight', schedule=my_sched)
    >>> net = Net(myModule, callbacks=[cb])

    Parameters
    ----------
    patterns : str or callable or list
      The pattern(s) to match parameter names against. Patterns are
      UNIX globbing patterns as understood by :func:`~fnmatch.fnmatch`.
      Patterns can also be callables which will get called with the
      parameter name and are regarded as a match when the callable
      returns a truthy value.

      This parameter also supports lists of str or callables so that
      one ``ParamMapper`` can match a group of parameters.

      Example: ``'linear*.weight'`` or ``['linear0.*', 'linear1.bias']``
      or ``lambda name: name.startswith('linear')``.

    fn : function
      The function to apply to each parameter separately.

    at : int or callable
      In case you specify an integer it represents the epoch number the
      function ``fn`` is applied to the parameters, in case ``at`` is
      a function it will receive ``net`` as parameter and the function
      is applied to the parameter once ``at`` returns ``True``.

    schedule : callable or None
      If specified this callable supersedes the static ``at``/``fn``
      combination by dynamically returning the function that is applied
      on the matched parameters. This way you can, for example, create a
      schedule that periodically freezes and unfreezes layers.

      The callable's signature is ``schedule(net: NeuralNet) -> callable``.

    """
    def __init__(self, patterns, fn=noop, at=1, schedule=None):
        self.at = at
        self.fn = fn
        self.schedule = schedule
        self.patterns = patterns

    def initialize(self):
        if not self.schedule:
            self.schedule = self._default_schedule

        if not isinstance(self.patterns, (list, tuple)):
            self.patterns = [self.patterns]

        if isinstance(self.at, int):
            if self.at <= 0:
                raise ValueError(
                    'Invalid value for `at` (at={}). The first possible '
                    'epoch number is 1.'.format(self.at))
            self.at = partial(self._epoch_at, epoch=self.at)

        return self

    def named_parameters(self, net):
        return net.get_all_learnable_params()

    def filter_parameters(self, patterns, params):
        pattern_fns = (
            pattern if callable(pattern) else partial(fnmatch, pat=pattern)
            for pattern in patterns
        )
        for pattern_fn, (name, param) in product(pattern_fns, params):
            if pattern_fn(name):
                yield name, param

    def _default_schedule(self, net):
        if self.at(net):
            return self.fn
        return noop

    def _epoch_at(self, net, epoch=1):
        return len(net.history) == epoch

    def on_epoch_begin(self, net, **kwargs):
        params = self.named_parameters(net)
        params = self.filter_parameters(self.patterns, params)
        map_fn = self.schedule(net)

        for _, p in params:
            map_fn(p)



class Freezer(ParamMapper):
    """Freeze matching parameters at the start of the first epoch. You may
    specify a specific point in time (either by epoch number or using a
    callable) when the parameters are frozen using the ``at`` parameter.

    See :class:`.ParamMapper` for details.
    """
    def __init__(self, *args, **kwargs):
        kwargs['at'] = kwargs.get('at', 1)
        kwargs['fn'] = kwargs.get('fn', freeze_parameter)
        super().__init__(*args, **kwargs)


class Unfreezer(ParamMapper):
    """Inverse operation of :class:`.Freezer`."""
    def __init__(self, *args, **kwargs):
        kwargs['at'] = kwargs.get('at', 1)
        kwargs['fn'] = kwargs.get('fn', unfreeze_parameter)
        super().__init__(*args, **kwargs)


class Initializer(ParamMapper):
    """Apply any function on matching parameters in the first epoch.

    Examples
    --------

    Use ``Initializer`` to initialize all dense layer weights with
    values sampled from an uniform distribution on the beginning of
    the first epoch:

    >>> init_fn = partial(torch.nn.init.uniform_, a=-1e-3, b=1e-3)
    >>> cb = Initializer('dense*.weight', fn=init_fn)
    >>> net = Net(myModule, callbacks=[cb])
    """
    def __init__(self, *args, **kwargs):
        kwargs['at'] = kwargs.get('at', 1)
        super().__init__(*args, **kwargs)


class LoadInitState(Callback):
    """Loads the model, optimizer, and history from a checkpoint into a
    :class:`.NeuralNet` when training begins.

    Examples
    --------

    Consider running the following example multiple times:

    >>> cp = Checkpoint(monitor='valid_loss_best')
    >>> load_state = LoadInitState(cp)
    >>> net = NeuralNet(..., callbacks=[cp, load_state])
    >>> net.fit(X, y)

    On the first run, the :class:`.Checkpoint` saves the model, optimizer, and
    history when the validation loss is minimized. During the first run,
    there are no files on disk, thus :class:`.LoadInitState` will
    not load anything. When running the example a second time,
    :class:`LoadInitState` will load the best model from the first run and
    continue training from there.

    Parameters
    ----------
    checkpoint: :class:`.Checkpoint`
      Checkpoint to get filenames from.

    """
    def __init__(self, checkpoint):
        self.checkpoint = checkpoint

    def initialize(self):
        self.did_load_ = False
        return self

    def on_train_begin(self, net,
                       X=None, y=None, **kwargs):
        if not self.did_load_:
            self.did_load_ = True
            with suppress(FileNotFoundError):
                if isinstance(self.checkpoint, TrainEndCheckpoint):
                    net.load_params(checkpoint=self.checkpoint.checkpoint_)
                else:
                    net.load_params(checkpoint=self.checkpoint)


class TrainEndCheckpoint(Callback):
    """Saves the model parameters, optimizer state, and history at the end of
    training. The default ``fn_prefix`` is ``'train_end_'``.

    Examples
    --------

    Consider running the following example multiple times:

    >>> train_end_cp = TrainEndCheckpoint(dirname='exp1')
    >>> load_state = LoadInitState(train_end_cp)
    >>> net = NeuralNet(..., callbacks=[train_end_cp, load_state])
    >>> net.fit(X, y)

    After the first run, model parameters, optimizer state, and history are
    saved into a directory named `exp1`. On the next run, `LoadInitState` will
    load the state from the first run and continue training.

    Parameters
    ----------
    f_params : file-like object, str, None (default='params.pt')
      File path to the file or file-like object where the model
      parameters should be saved. Pass ``None`` to disable saving
      model parameters.

      If the value is a string you can also use format specifiers
      to, for example, indicate the current epoch. Accessible format
      values are ``net``, ``last_epoch`` and ``last_batch``.
      Example to include last epoch number in file name:

      >>> cb = Checkpoint(f_params="params_{last_epoch[epoch]}.pt")

    f_optimizer : file-like object, str, None (default='optimizer.pt')
      File path to the file or file-like object where the optimizer
      state should be saved. Pass ``None`` to disable saving
      model parameters.

      Supports the same format specifiers as ``f_params``.

    f_criterion : file-like object, str, None (default='criterion.pt')
      File path to the file or file-like object where the criterion
      state should be saved. Pass ``None`` to disable saving
      model parameters.

      Supports the same format specifiers as ``f_params``.

    f_history : file-like object, str, None (default='history.json')
      File path to the file or file-like object where the model
      training history should be saved. Pass ``None`` to disable
      saving history.

    f_pickle : file-like object, str, None (default=None)
      File path to the file or file-like object where the entire
      model object should be pickled. Pass ``None`` to disable
      pickling.

      Supports the same format specifiers as ``f_params``.

    fn_prefix: str (default='train_end_')
      Prefix for filenames. If ``f_params``, ``f_optimizer``, ``f_history``,
      or ``f_pickle`` are strings, they will be prefixed by ``fn_prefix``.

    dirname: str (default='')
      Directory where files are stored.

    sink : callable (default=noop)
      The target that the information about created checkpoints is
      sent to. This can be a logger or ``print`` function (to send to
      stdout). By default the output is discarded.

    """
    def __init__(
            self,
            f_params='params.pt',
            f_optimizer='optimizer.pt',
            f_criterion='criterion.pt',
            f_history='history.json',
            f_pickle=None,
            fn_prefix='train_end_',
            dirname='',
            sink=noop,
            **kwargs
    ):
        self.f_params = f_params
        self.f_optimizer = f_optimizer
        self.f_criterion = f_criterion
        self.f_history = f_history
        self.f_pickle = f_pickle
        self.fn_prefix = fn_prefix
        self.dirname = dirname
        self.sink = sink
        Checkpoint._check_kwargs(self, kwargs)
        vars(self).update(**kwargs)

    def _f_kwargs(self):
        return {name: getattr(self, name) for name in dir(self)
                if name.startswith('f_')}

    def initialize(self):
        self.checkpoint_ = Checkpoint(
            monitor=None,
            fn_prefix=self.fn_prefix,
            dirname=self.dirname,
            event_name=None,
            sink=self.sink,
            **self._f_kwargs()
        )
        self.checkpoint_.initialize()
        return self

    def on_train_end(self, net, **kwargs):
        self.checkpoint_.save_model(net)
        self.checkpoint_._sink("Final checkpoint triggered", net.verbose)
        return self


class InputShapeSetter(Callback):
    """Sets the input dimension of the PyTorch module to the input dimension
    of the training data. By default the last dimension of X (``X.shape[-1]``)
    will be used.

    This can be of use when the shape of X is not known beforehand,
    e.g. when using a skorch model within an sklearn pipeline and
    grid-searching feature transformers, or using feature selection
    methods.

    Basic usage:

    >>> class MyModule(torch.nn.Module):
    ...     def __init__(self, input_dim=1):
    ...         super().__init__()
    ...         self.layer = torch.nn.Linear(input_dim, 3)
    ... # ...
    >>> X1 = np.zeros(100, 5)
    >>> X2 = np.zeros(100, 3)
    >>> y = np.zeros(100)
    >>> net = NeuralNetClassifier(MyModule, callbacks=[InputShapeSetter()])
    >>> net.fit(X1, y)  # self.module_.layer.in_features == 5
    >>> net.fit(X2, y)  # self.module_.layer.in_features == 3

    Parameters
    ----------
    param_name : str (default='input_dim')
      The parameter name is the parameter your model uses to define the
      input dimension in its ``__init__`` method.

    input_dim_fn : callable, None (default=None)
      In case your ``X`` value is more complex and deriving the input
      dimension is not as easy as ``X.shape[-1]`` you can pass a callable
      to this parameter which takes ``X`` and returns the input dimension.

    module_name : str (default='module')
      Only needs change when you are using more than one module in your
      skorch model (e.g., in case of GANs).
    """
    def __init__(
        self,
        param_name='input_dim',
        input_dim_fn=None,
        module_name='module',
    ):
        self.module_name = module_name
        self.param_name = param_name
        self.input_dim_fn = input_dim_fn

    def get_input_dim(self, X):
        if self.input_dim_fn is not None:
            return self.input_dim_fn(X)
        if len(X.shape) < 2:
            raise ValueError(
                "Expected at least two-dimensional input data for X. "
                "If your data is one-dimensional, please use the "
                "`input_dim_fn` parameter to infer the correct "
                "input shape."
            )
        return X.shape[-1]

    def on_train_begin(self, net, X, y, **kwargs):
        params = net.get_params()
        input_dim = self.get_input_dim(X)
        param_name = f'{self.module_name}__{self.param_name}'

        if params.get(param_name, None) == input_dim:
            return

        kwargs = {param_name: input_dim}
        net.set_params(**kwargs)



""" Basic callback definition. """

import warnings

from sklearn.base import BaseEstimator
from skorch.exceptions import SkorchWarning


__all__ = ['Callback']


class Callback:
    """Base class for callbacks.

    All custom callbacks should inherit from this class. The subclass
    may override any of the ``on_...`` methods. It is, however, not
    necessary to override all of them, since it's okay if they don't
    have any effect.

    Classes that inherit from this also gain the ``get_params`` and
    ``set_params`` method.

    """
    def initialize(self):
        """(Re-)Set the initial state of the callback. Use this
        e.g. if the callback tracks some state that should be reset
        when the model is re-initialized.

        This method should return self.

        """
        return self

    def on_train_begin(self, net, X=None, y=None, **kwargs):
        """Called at the beginning of training."""

    def on_train_end(self, net, X=None, y=None, **kwargs):
        """Called at the end of training."""

    def on_epoch_begin(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        """Called at the beginning of each epoch."""

    def on_epoch_end(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        """Called at the end of each epoch."""

    def on_batch_begin(self, net, batch=None, training=None, **kwargs):
        """Called at the beginning of each batch."""

    def on_batch_end(self, net, batch=None, training=None, **kwargs):
        """Called at the end of each batch."""

    def on_grad_computed(
            self, net, named_parameters, X=None, y=None, training=None, **kwargs):
        """Called once per batch after gradients have been computed but before
        an update step was performed.
        """

    def _get_param_names(self):
        return (key for key in self.__dict__ if not key.endswith('_'))

    def get_params(self, deep=True):
        return BaseEstimator.get_params(self, deep=deep)

    def set_params(self, **params):
        BaseEstimator.set_params(self, **params)



""" Callbacks for calculating scores."""

from contextlib import contextmanager
from contextlib import suppress
from functools import partial
import warnings

import numpy as np
import sklearn
from sklearn.metrics import make_scorer, check_scoring

from skorch.callbacks import Callback
from skorch.dataset import unpack_data
from sklearn.metrics._scorer import _BaseScorer
from skorch.utils import data_from_dataset
from skorch.utils import is_skorch_dataset
from skorch.utils import to_numpy
from skorch.utils import check_indexing
from skorch.utils import to_device


__all__ = ['BatchScoring', 'EpochScoring', 'PassthroughScoring']


@contextmanager
def cache_net_infer(net, use_caching, y_preds):
    """Caching context for ``skorch.NeuralNet`` instance.

    Returns a modified version of the net whose ``infer`` method will
    subsequently return cached predictions. Leaving the context will
    undo the overwrite of the ``infer`` method.

    Deprecated.

    """

    warnings.warn(
        "cache_net_infer is no longer uesd to provide caching for "
        "the scoring callbacks and will hence be removed in a "
        "future release.",
        DeprecationWarning,
    )

    if not use_caching:
        yield net
        return
    y_preds = iter(y_preds)
    net.infer = lambda *a, **kw: next(y_preds)

    try:
        yield net
    finally:
        # By setting net.infer we define an attribute `infer`
        # that precedes the bound method `infer`. By deleting
        # the entry from the attribute dict we undo this.
        del net.__dict__['infer']


@contextmanager
def _cache_net_forward_iter(net, use_caching, y_preds):
    """Caching context for ``skorch.NeuralNet`` instance.

    Returns a modified version of the net whose ``forward_iter``
    method will subsequently return cached predictions. Leaving the
    context will undo the overwrite of the ``forward_iter`` method.

    """
    if not use_caching:
        yield net
        return
    y_preds = iter(y_preds)

    # pylint: disable=unused-argument
    def cached_forward_iter(*args, device=net.device, **kwargs):
        for yp in y_preds:
            yield to_device(yp, device=device)

    net.forward_iter = cached_forward_iter
    try:
        yield net
    finally:
        # By setting net.forward_iter we define an attribute
        # `forward_iter` that precedes the bound method
        # `forward_iter`. By deleting the entry from the attribute
        # dict we undo this.
        del net.__dict__['forward_iter']


def convert_sklearn_metric_function(scoring):
    """If ``scoring`` is a sklearn metric function, convert it to a
    sklearn scorer and return it. Otherwise, return ``scoring`` unchanged."""
    if callable(scoring):
        module = getattr(scoring, '__module__', None)

        # those are scoring objects returned by make_scorer starting
        # from sklearn 0.22
        scorer_names = ('_PredictScorer', '_ProbaScorer', '_ThresholdScorer')
        if (
                hasattr(module, 'startswith') and
                module.startswith('sklearn.metrics.') and
                not module.startswith('sklearn.metrics.scorer') and
                not module.startswith('sklearn.metrics.tests.') and
                not scoring.__class__.__name__ in scorer_names
        ):
            return make_scorer(scoring)
    return scoring


class ScoringBase(Callback):
    """Base class for scoring.

    Subclass and implement an ``on_*`` method before using.
    """
    def __init__(
            self,
            scoring,
            lower_is_better=True,
            on_train=False,
            name=None,
            target_extractor=to_numpy,
            use_caching=True,
    ):
        self.scoring = scoring
        self.lower_is_better = lower_is_better
        self.on_train = on_train
        self.name = name
        self.target_extractor = target_extractor
        self.use_caching = use_caching

    # pylint: disable=protected-access
    def _get_name(self):
        """Find name of scoring function."""
        if self.name is not None:
            return self.name
        if self.scoring_ is None:
            return 'score'
        if isinstance(self.scoring_, str):
            return self.scoring_
        if isinstance(self.scoring_, partial):
            return self.scoring_.func.__name__
        if isinstance(self.scoring_, _BaseScorer):
            if hasattr(self.scoring_._score_func, '__name__'):
                # sklearn < 0.22
                return self.scoring_._score_func.__name__
            # sklearn >= 0.22
            return self.scoring_._score_func._score_func.__name__
        if isinstance(self.scoring_, dict):
            raise ValueError("Dict not supported as scorer for multi-metric scoring."
                             " Register multiple scoring callbacks instead.")
        return self.scoring_.__name__

    def initialize(self):
        self.best_score_ = np.inf if self.lower_is_better else -np.inf
        self.scoring_ = convert_sklearn_metric_function(self.scoring)
        self.name_ = self._get_name()
        return self

    # pylint: disable=attribute-defined-outside-init,arguments-differ
    def on_train_begin(self, net, X, y, **kwargs):
        self.X_indexing_ = check_indexing(X)
        self.y_indexing_ = check_indexing(y)

        # Looks for the right most index where `*_best` is True
        # That index is used to get the best score in `net.history`
        with suppress(ValueError, IndexError, KeyError):
            best_name_history = net.history[:, '{}_best'.format(self.name_)]
            idx_best_reverse = best_name_history[::-1].index(True)
            idx_best = len(best_name_history) - idx_best_reverse - 1
            self.best_score_ = net.history[idx_best, self.name_]

    def _scoring(self, net, X_test, y_test):
        """Resolve scoring and apply it to data. Use cached prediction
        instead of running inference again, if available."""
        scorer = check_scoring(net, self.scoring_)
        return scorer(net, X_test, y_test)

    def _is_best_score(self, current_score):
        if self.lower_is_better is None:
            return None
        if self.lower_is_better:
            return current_score < self.best_score_
        return current_score > self.best_score_


class BatchScoring(ScoringBase):
    """Callback that performs generic scoring on batches.

    This callback determines the score after each batch and stores it
    in the net's history in the column given by ``name``. At the end
    of the epoch, the average of the scores are determined and also
    stored in the history. Furthermore, it is determined whether this
    average score is the best score yet and that information is also
    stored in the history.

    In contrast to :class:`.EpochScoring`, this callback determines
    the score for each batch and then averages the score at the end of
    the epoch. This can be disadvantageous for some scores if the
    batch size is small -- e.g. area under the ROC will return
    incorrect scores in this case. Therefore, it is recommnded to use
    :class:`.EpochScoring` unless you really need the scores for each
    batch.

    If ``y`` is None, the ``scoring`` function with signature (model, X, y)
    must be able to handle ``X`` as a ``Tensor`` and ``y=None``.

    Parameters
    ----------
    scoring : None, str, or callable
      If None, use the ``score`` method of the model. If str, it should
      be a valid sklearn metric (e.g. "f1_score", "accuracy_score"). If
      a callable, it should have the signature (model, X, y), and it
      should return a scalar. This works analogously to the ``scoring``
      parameter in sklearn's ``GridSearchCV`` et al.

    lower_is_better : bool (default=True)
      Whether lower (e.g. log loss) or higher (e.g. accuracy) scores
      are better.

    on_train : bool (default=False)
      Whether this should be called during train or validation.

    name : str or None (default=None)
      If not an explicit string, tries to infer the name from the
      ``scoring`` argument.

    target_extractor : callable (default=to_numpy)
      This is called on y before it is passed to scoring.

    use_caching : bool (default=True)
      Re-use the model's prediction for computing the loss to calculate
      the score. Turning this off will result in an additional inference
      step for each batch.

    """
    # pylint: disable=unused-argument,arguments-differ

    def on_batch_end(self, net, batch, training, **kwargs):
        if training != self.on_train:
            return

        X, y = unpack_data(batch)
        y_preds = [kwargs['y_pred']]
        with _cache_net_forward_iter(net, self.use_caching, y_preds) as cached_net:
            # In case of y=None we will not have gathered any samples.
            # We expect the scoring function to deal with y=None.
            y = None if y is None else self.target_extractor(y)
            try:
                score = self._scoring(cached_net, X, y)
                cached_net.history.record_batch(self.name_, score)
            except KeyError:
                pass

    def get_avg_score(self, history):
        if self.on_train:
            bs_key = 'train_batch_size'
        else:
            bs_key = 'valid_batch_size'

        weights, scores = list(zip(
            *history[-1, 'batches', :, [bs_key, self.name_]]))
        score_avg = np.average(scores, weights=weights)
        return score_avg

    # pylint: disable=unused-argument
    def on_epoch_end(self, net, **kwargs):
        history = net.history
        try:  # don't raise if there is no valid data
            history[-1, 'batches', :, self.name_]
        except KeyError:
            return

        score_avg = self.get_avg_score(history)
        is_best = self._is_best_score(score_avg)
        if is_best:
            self.best_score_ = score_avg

        history.record(self.name_, score_avg)
        if is_best is not None:
            history.record(self.name_ + '_best', bool(is_best))


class EpochScoring(ScoringBase):
    """Callback that performs generic scoring on predictions.

    At the end of each epoch, this callback makes a prediction on
    train or validation data, determines the score for that prediction
    and whether it is the best yet, and stores the result in the net's
    history.

    In case you already computed a score value for each batch you
    can omit the score computation step by return the value from
    the history. For example:

        >>> def my_score(net, X=None, y=None):
        ...     losses = net.history[-1, 'batches', :, 'my_score']
        ...     batch_sizes = net.history[-1, 'batches', :, 'valid_batch_size']
        ...     return np.average(losses, weights=batch_sizes)
        >>> net = MyNet(callbacks=[
        ...     ('my_score', Scoring(my_score, name='my_score'))

    If you fit with a custom dataset, this callback should work as
    expected as long as ``use_caching=True`` which enables the
    collection of ``y`` values from the dataset. If you decide to
    disable the caching of predictions and ``y`` values, you need
    to write your own scoring function that is able to deal with the
    dataset and returns a scalar, for example:

        >>> def ds_accuracy(net, ds, y=None):
        ...     # assume ds yields (X, y), e.g. torchvision.datasets.MNIST
        ...     y_true = [y for _, y in ds]
        ...     y_pred = net.predict(ds)
        ...     return sklearn.metrics.accuracy_score(y_true, y_pred)
        >>> net = MyNet(callbacks=[
        ...     EpochScoring(ds_accuracy, use_caching=False)])
        >>> ds = torchvision.datasets.MNIST(root=mnist_path)
        >>> net.fit(ds)

    Parameters
    ----------
    scoring : None, str, or callable (default=None)
      If None, use the ``score`` method of the model. If str, it
      should be a valid sklearn scorer (e.g. "f1", "accuracy"). If a
      callable, it should have the signature (model, X, y), and it
      should return a scalar. This works analogously to the
      ``scoring`` parameter in sklearn's ``GridSearchCV`` et al.

    lower_is_better : bool (default=True)
      Whether lower scores should be considered better or worse.

    on_train : bool (default=False)
      Whether this should be called during train or validation data.

    name : str or None (default=None)
      If not an explicit string, tries to infer the name from the
      ``scoring`` argument.

    target_extractor : callable (default=to_numpy)
      This is called on y before it is passed to scoring.

    use_caching : bool (default=True)
      Collect labels and predictions (``y_true`` and ``y_pred``)
      over the course of one epoch and use the cached values for
      computing the score. The cached values are shared between
      all ``EpochScoring`` instances. Disabling this will result
      in an additional inference step for each epoch and an
      inability to use arbitrary datasets as input (since we
      don't know how to extract ``y_true`` from an arbitrary
      dataset).

    """
    def _initialize_cache(self):
        self.y_trues_ = []
        self.y_preds_ = []

    def initialize(self):
        super().initialize()
        self._initialize_cache()
        return self

    # pylint: disable=arguments-differ,unused-argument
    def on_epoch_begin(self, net, dataset_train, dataset_valid, **kwargs):
        self._initialize_cache()

    # pylint: disable=arguments-differ
    def on_batch_end(
            self, net, batch, y_pred, training, **kwargs):
        if not self.use_caching or training != self.on_train:
            return

        # We collect references to the prediction and target data
        # emitted by the training process. Since we don't copy the
        # data, all *Scoring callback instances use the same
        # underlying data. This is also the reason why we don't run
        # self.target_extractor(y) here but on epoch end, so that
        # there are no copies of parts of y hanging around during
        # training.
        _X, y = unpack_data(batch)
        if y is not None:
            self.y_trues_.append(y)
        self.y_preds_.append(y_pred)

    def get_test_data(self, dataset_train, dataset_valid):
        """Return data needed to perform scoring.

        This is a convenience method that handles picking of
        train/valid, different types of input data, use of cache,
        etc. for you.

        Parameters
        ----------
        dataset_train
          Incoming training data or dataset.

        dataset_valid
          Incoming validation data or dataset.

        Returns
        -------
        X_test
          Input data used for making the prediction.

        y_test
          Target ground truth. If caching was enabled, return cached
          y_test.

        y_pred : list
          The predicted targets. If caching was disabled, the list is
          empty. If caching was enabled, the list contains the batches
          of the predictions. It may thus be necessary to concatenate
          the output before working with it:
          ``y_pred = np.concatenate(y_pred)``

        """
        dataset = dataset_train if self.on_train else dataset_valid

        if self.use_caching:
            X_test = dataset
            y_pred = self.y_preds_
            y_test = [self.target_extractor(y) for y in self.y_trues_]
            # In case of y=None we will not have gathered any samples.
            # We expect the scoring function to deal with y_test=None.
            y_test = np.concatenate(y_test) if y_test else None
            return X_test, y_test, y_pred

        if is_skorch_dataset(dataset):
            X_test, y_test = data_from_dataset(
                dataset,
                X_indexing=self.X_indexing_,
                y_indexing=self.y_indexing_,
            )
        else:
            X_test, y_test = dataset, None

        if y_test is not None:
            # We allow y_test to be None but the scoring function has
            # to be able to deal with it (i.e. called without y_test).
            y_test = self.target_extractor(y_test)
        return X_test, y_test, []

    def _record_score(self, history, current_score):
        """Record the current store and, if applicable, if it's the best score
        yet.

        """
        history.record(self.name_, current_score)

        is_best = self._is_best_score(current_score)
        if is_best is None:
            return

        history.record(self.name_ + '_best', bool(is_best))
        if is_best:
            self.best_score_ = current_score

    # pylint: disable=unused-argument,arguments-differ
    def on_epoch_end(
            self,
            net,
            dataset_train,
            dataset_valid,
            **kwargs):
        X_test, y_test, y_pred = self.get_test_data(dataset_train, dataset_valid)
        if X_test is None:
            return

        with _cache_net_forward_iter(net, self.use_caching, y_pred) as cached_net:
            current_score = self._scoring(cached_net, X_test, y_test)

        self._record_score(net.history, current_score)

    def on_train_end(self, *args, **kwargs):
        self._initialize_cache()


class PassthroughScoring(Callback):
    """Creates scores on epoch level based on batch level scores

    This callback doesn't calculate any new scores but instead passes
    through a score that was created on the batch level. Based on that
    score, an average across the batch is created (honoring the batch
    size) and recorded in the history for the given epoch.

    Use this callback when there already is a score calculated on the
    batch level. If that score has yet to be calculated, use
    :class:`.BatchScoring` instead.

    Parameters
    ----------
    name : str
      Name of the score recorded on a batch level in the history.

    lower_is_better : bool (default=True)
      Whether lower (e.g. log loss) or higher (e.g. accuracy) scores
      are better.

    on_train : bool (default=False)
      Whether this should be called during train or validation.

    """
    def __init__(
            self,
            name,
            lower_is_better=True,
            on_train=False,
    ):
        self.name = name
        self.lower_is_better = lower_is_better
        self.on_train = on_train

    def initialize(self):
        self.best_score_ = np.inf if self.lower_is_better else -np.inf
        return self

    def _is_best_score(self, current_score):
        if self.lower_is_better is None:
            return None
        if self.lower_is_better:
            return current_score < self.best_score_
        return current_score > self.best_score_

    def get_avg_score(self, history):
        if self.on_train:
            bs_key = 'train_batch_size'
        else:
            bs_key = 'valid_batch_size'

        weights, scores = list(zip(
            *history[-1, 'batches', :, [bs_key, self.name]]))
        score_avg = np.average(scores, weights=weights)
        return score_avg

    # pylint: disable=unused-argument,arguments-differ
    def on_epoch_end(self, net, **kwargs):
        history = net.history
        try:  # don't raise if there is no valid data
            history[-1, 'batches', :, self.name]
        except KeyError:
            return

        score_avg = self.get_avg_score(history)
        is_best = self._is_best_score(score_avg)
        if is_best:
            self.best_score_ = score_avg

        history.record(self.name, score_avg)
        if is_best is not None:
            history.record(self.name + '_best', bool(is_best))



"""Tests for toy.py."""

import numpy as np
import pytest
import torch
from torch import nn


class TestMLPModule:
    @pytest.fixture
    def module_cls(self):
        from skorch.toy import MLPModule
        return MLPModule

    def test_one_hidden(self, module_cls):
        module = module_cls()
        parameters = list(module.named_parameters())

        # 2 linear * (weight + bias)
        assert len(parameters) == 4

        # 2 linear, 1 relu, 1 dropout
        assert len(module.sequential) == 4
        assert isinstance(module.sequential[0], nn.Linear)
        assert isinstance(module.sequential[1], nn.ReLU)
        assert isinstance(module.sequential[2], nn.Dropout)
        assert isinstance(module.sequential[3], nn.Linear)

    def test_two_hidden(self, module_cls):
        module = module_cls(num_hidden=2)
        parameters = list(module.named_parameters())

        # 3 linear * (weight + bias)
        assert len(parameters) == 6

        # 3 linear, 2 relu, 2 dropout
        assert len(module.sequential) == 7
        assert isinstance(module.sequential[0], nn.Linear)
        assert isinstance(module.sequential[1], nn.ReLU)
        assert isinstance(module.sequential[2], nn.Dropout)
        assert isinstance(module.sequential[3], nn.Linear)
        assert isinstance(module.sequential[4], nn.ReLU)
        assert isinstance(module.sequential[5], nn.Dropout)
        assert isinstance(module.sequential[6], nn.Linear)

    @pytest.mark.parametrize('num_hidden', [0, 1, 2, 5, 10])
    def test_many_hidden(self, module_cls, num_hidden):
        module = module_cls(num_hidden=num_hidden)
        parameters = list(module.named_parameters())

        assert len(parameters) == 2 * (num_hidden + 1)
        assert len(module.sequential) == (3 * num_hidden) + 1

    def test_output_nonlin(self, module_cls):
        module = module_cls(output_nonlin=nn.Sigmoid())

        # 2 linear, 1 relu, 1 dropout, 1 sigmoid
        assert len(module.sequential) == 5
        assert isinstance(module.sequential[0], nn.Linear)
        assert isinstance(module.sequential[1], nn.ReLU)
        assert isinstance(module.sequential[2], nn.Dropout)
        assert isinstance(module.sequential[3], nn.Linear)
        assert isinstance(module.sequential[4], nn.Sigmoid)

    def test_output_squeezed(self, module_cls):
        X = torch.zeros((5, 20)).float()

        module = module_cls(output_units=1)
        y = module(X)
        assert y.dim() == 2

        module = module_cls(squeeze_output=True, output_units=1)
        y = module(X)
        assert y.dim() == 1

    def test_dropout(self, module_cls):
        module = module_cls(dropout=0.567)
        assert np.isclose(module.sequential[2].p, 0.567)

    def test_make_classifier(self):
        from skorch.toy import make_classifier
        module = make_classifier()()
        assert isinstance(module.sequential[-1], nn.Softmax)

    def test_make_binary_classifier(self):
        from skorch.toy import make_binary_classifier
        module = make_binary_classifier()()
        assert isinstance(module.sequential[-1], nn.Linear)
        assert module.squeeze_output is True

    def test_make_regressor(self):
        from skorch.toy import make_regressor
        module = make_regressor()()
        assert module.sequential[-1].out_features == 1



"""Test for cli.py"""

from math import cos
from unittest.mock import Mock
from unittest.mock import patch

import numpy as np
import pytest
from sklearn.pipeline import FeatureUnion
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from torch import nn
from torch.nn import RReLU


fire_installed = True
try:
    import fire  # pylint: disable=unused-import
except ImportError:
    fire_installed = False


@pytest.mark.skipif(not fire_installed, reason='fire libarary not installed')
class TestCli:
    @pytest.fixture
    def resolve_dotted_name(self):
        from skorch.cli import _resolve_dotted_name
        return _resolve_dotted_name

    @pytest.mark.parametrize('name, expected', [
        (0, 0),
        (1.23, 1.23),
        ('foo', 'foo'),
        ('math.cos', cos),
        ('torch.nn', nn),
        ('torch.nn.ReLU', nn.ReLU),
    ])
    def test_resolve_dotted_name(self, resolve_dotted_name, name, expected):
        result = resolve_dotted_name(name)
        assert result == expected

    def test_resolve_dotted_name_instantiated(self, resolve_dotted_name):
        result = resolve_dotted_name('torch.nn.RReLU(0.123, upper=0.456)')
        assert isinstance(result, RReLU)
        assert np.isclose(result.lower, 0.123)
        assert np.isclose(result.upper, 0.456)

    @pytest.fixture
    def parse_net_kwargs(self):
        from skorch.cli import parse_net_kwargs
        return parse_net_kwargs

    def test_parse_net_kwargs(self, parse_net_kwargs):
        kwargs = {
            'lr': 0.05,
            'max_epochs': 5,
            'module__num_units': 10,
            'module__nonlin': 'torch.nn.RReLU(0.123, upper=0.456)',
        }
        parsed_kwargs = parse_net_kwargs(kwargs)

        assert len(parsed_kwargs) == 4
        assert np.isclose(parsed_kwargs['lr'], 0.05)
        assert parsed_kwargs['max_epochs'] == 5
        assert parsed_kwargs['module__num_units'] == 10
        assert isinstance(parsed_kwargs['module__nonlin'], RReLU)
        assert np.isclose(parsed_kwargs['module__nonlin'].lower, 0.123)
        assert np.isclose(parsed_kwargs['module__nonlin'].upper, 0.456)

    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def net(self, net_cls, classifier_module):
        return net_cls(classifier_module)

    @pytest.fixture
    def pipe(self, net):
        return Pipeline([
            ('features', FeatureUnion([
                ('scale', MinMaxScaler()),
            ])),
            ('net', net),
        ])

    @pytest.fixture
    def yield_estimators(self):
        from skorch.cli import _yield_estimators
        return _yield_estimators

    def test_yield_estimators_net(self, yield_estimators, net):
        result = list(yield_estimators(net))

        assert result[0][0] == ''
        assert result[0][1] is net
        assert result[1][0] == 'module'
        assert result[1][1] is net.module

    def test_yield_estimators_pipe(self, yield_estimators, pipe):
        result = list(yield_estimators(pipe))
        scaler = pipe.named_steps['features'].transformer_list[0][1]
        net = pipe.named_steps['net']
        module = net.module

        assert result[0][0] == 'features__scale'
        assert result[0][1] is scaler
        assert result[1][0] == 'net'
        assert result[1][1] is net
        assert result[2][0] == 'net__module'
        assert result[2][1] is module

    @pytest.fixture
    def substitute_default(self):
        from skorch.cli import _substitute_default
        return _substitute_default

    @pytest.mark.parametrize('s, new_value, expected', [
        ('', '', ''),
        ('', 'foo', ''),
        ('bar', 'foo', 'bar'),
        ('int (default=128)', '', 'int (default=)'),
        ('int (default=128)', None, 'int (default=128)'),
        ('int (default=128)', '""', 'int (default="")'),
        ('int (default=128)', '128', 'int (default=128)'),
        ('int (default=128)', '256', 'int (default=256)'),
        ('int (default=128)', 256, 'int (default=256)'),
        ('with_parens (default=(1, 2))', (3, 4), 'with_parens (default=(3, 4))'),
        ('int (default =128)', '256', 'int (default =256)'),
        ('int (default= 128)', '256', 'int (default= 256)'),
        ('int (default = 128)', '256', 'int (default = 256)'),
        (
            'nonlin (default = ReLU())',
            nn.Hardtanh(1, 2),
            'nonlin (default = {})'.format(nn.Hardtanh(1, 2))
        ),
        (
            # from sklearn MinMaxScaler
            'tuple (min, max), default=(0, 1)',
            (-1, 1),
            'tuple (min, max), default=(-1, 1)'
        ),
        (
            # from sklearn MinMaxScaler
            'boolean, optional, default True',
            False,
            'boolean, optional, default False'
        ),
        (
            # from sklearn Normalizer
            "'l1', 'l2', or 'max', optional ('l2' by default)",
            'l1',
            "'l1', 'l2', or 'max', optional ('l1' by default)"
        ),
        (
            # same but double ticks
            '"l1", "l2", or "max", optional ("l2" by default)',
            'l1',
            '"l1", "l2", or "max", optional ("l1" by default)'
        ),
        (
            # same but no ticks
            "l1, l2, or max, optional (l2 by default)",
            'l1',
            "l1, l2, or max, optional (l1 by default)"
        ),
        (
            "tuple, optional ((1, 1) by default)",
            (2, 2),
            "tuple, optional ((2, 2) by default)"
        ),
        (
            "nonlin (ReLU() by default)",
            nn.Tanh(),
            "nonlin (Tanh() by default)"
        ),
    ])
    def test_replace_default(self, substitute_default, s, new_value, expected):
        result = substitute_default(s, new_value)
        assert result == expected

    @pytest.fixture
    def print_help(self):
        from skorch.cli import print_help
        return print_help

    def test_print_help_net(self, print_help, net, capsys):
        print_help(net)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '<NeuralNetClassifier> options',
            '--module : torch module (class or instance)',
            '--batch_size : int (default=128)',
            '<MLPModule> options',
            '--module__hidden_units : int (default=10)'
        ]
        for snippet in expected_snippets:
            assert snippet in out

    def test_print_help_net_custom_defaults(self, print_help, net, capsys):
        defaults = {'batch_size': 256, 'module__hidden_units': 55}
        print_help(net, defaults)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '<NeuralNetClassifier> options',
            '--module : torch module (class or instance)',
            '--batch_size : int (default=256)',
            '<MLPModule> options',
            '--module__hidden_units : int (default=55)'
        ]
        for snippet in expected_snippets:
            assert snippet in out

    def test_print_help_pipeline(self, print_help, pipe, capsys):
        print_help(pipe)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '<MinMaxScaler> options',
            '--features__scale__feature_range',
            '<NeuralNetClassifier> options',
            '--net__module : torch module (class or instance)',
            '--net__batch_size : int (default=128)',
            '<MLPModule> options',
            '--net__module__hidden_units : int (default=10)'
        ]
        for snippet in expected_snippets:
            assert snippet in out

    def test_print_help_pipeline_custom_defaults(
            self, print_help, pipe, capsys):
        defaults = {'net__batch_size': 256, 'net__module__hidden_units': 55}
        print_help(pipe, defaults=defaults)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '<MinMaxScaler> options',
            '--features__scale__feature_range',
            '<NeuralNetClassifier> options',
            '--net__module : torch module (class or instance)',
            '--net__batch_size : int (default=256)',
            '<MLPModule> options',
            '--net__module__hidden_units : int (default=55)'
        ]
        for snippet in expected_snippets:
            assert snippet in out

    @pytest.fixture
    def clf_sklearn(self):
        from sklearn.linear_model import LinearRegression
        return LinearRegression()

    @pytest.fixture
    def pipe_sklearn(self, clf_sklearn):
        return Pipeline([
            ('features', FeatureUnion([
                ('scale', MinMaxScaler()),
            ])),
            ('clf', clf_sklearn),
        ])

    def test_print_help_sklearn_estimator(self, print_help, clf_sklearn, capsys):
        # Should also work with non-skorch sklearn estimator;
        # need to assert that count==1 since there was a bug in my
        # first implementation that resulted in the help for the final
        # estimator appearing twice.
        print_help(clf_sklearn)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '--fit_intercept',
            '--copy_X',
            '--normalize',
        ]
        for snippet in expected_snippets:
            assert snippet in out
            assert out.count(snippet) == 1

    def test_print_help_sklearn_pipeline(self, print_help, pipe_sklearn, capsys):
        # Should also work with non-skorch sklearn pipelines;
        # need to assert that count==1 since there was a bug in my
        # first implementation that resulted in the help for the final
        # estimator appearing twice.
        print_help(pipe_sklearn)
        out = capsys.readouterr()[0]

        expected_snippets = [
            '-- --help',
            '<MinMaxScaler> options',
            '--features__scale__feature_range',
            '--clf__fit_intercept',
            '--clf__normalize',
        ]
        for snippet in expected_snippets:
            assert snippet in out
            assert out.count(snippet) == 1

    @pytest.fixture
    def parse_args(self):
        from skorch.cli import parse_args
        return parse_args

    @pytest.fixture
    def estimator(self, net_cls):
        mock = Mock(net_cls)
        return mock

    def test_parse_args_help(self, parse_args, estimator):
        with patch('skorch.cli.sys.exit') as exit_:
            with patch('skorch.cli.print_help') as help_:
                parsed = parse_args({'help': True, 'foo': 'bar'})
                parsed(estimator)

        assert estimator.set_params.call_count == 0  # kwargs and defaults
        assert help_.call_count == 1
        assert exit_.call_count == 1

    def test_parse_args_run(self, parse_args, estimator):
        kwargs = {'foo': 'bar', 'baz': 'math.cos'}
        with patch('skorch.cli.sys.exit') as exit_:
            with patch('skorch.cli.print_help') as help_:
                parsed = parse_args(kwargs)
                parsed(estimator)

        assert estimator.set_params.call_count == 2  # defaults and kwargs

        defaults_set_params = estimator.set_params.call_args_list[0][1]
        assert not defaults_set_params  # no defaults specified

        kwargs_set_params = estimator.set_params.call_args_list[1][1]
        assert kwargs_set_params['foo'] == 'bar'
        # pylint: disable=comparison-with-callable
        assert kwargs_set_params['baz'] == cos

        assert help_.call_count == 0
        assert exit_.call_count == 0

    def test_parse_args_net_custom_defaults(self, parse_args, net):
        defaults = {'batch_size': 256, 'module__hidden_units': 55}
        kwargs = {'batch_size': 123, 'module__nonlin': nn.Hardtanh(1, 2)}
        parsed = parse_args(kwargs, defaults)
        net = parsed(net)

        # cmd line args have precedence over defaults
        assert net.batch_size == 123
        assert net.module__hidden_units == 55
        assert isinstance(net.module__nonlin, nn.Hardtanh)
        assert net.module__nonlin.min_val == 1
        assert net.module__nonlin.max_val == 2

    def test_parse_args_pipe_custom_defaults(self, parse_args, pipe):
        defaults = {'net__batch_size': 256, 'net__module__hidden_units': 55}
        kwargs = {'net__batch_size': 123, 'net__module__nonlin': nn.Hardtanh(1, 2)}
        parsed = parse_args(kwargs, defaults)
        pipe = parsed(pipe)
        net = pipe.steps[-1][1]

        # cmd line args have precedence over defaults
        assert net.batch_size == 123
        assert net.module__hidden_units == 55
        assert isinstance(net.module__nonlin, nn.Hardtanh)
        assert net.module__nonlin.min_val == 1
        assert net.module__nonlin.max_val == 2

    def test_parse_args_sklearn_pipe_custom_defaults(self, parse_args, pipe_sklearn):
        defaults = {'features__scale__copy': 123, 'clf__fit_intercept': 456}
        kwargs = {'features__scale__copy': 555, 'clf__normalize': 789}
        parsed = parse_args(kwargs, defaults)
        pipe = parsed(pipe_sklearn)
        scaler = pipe.steps[0][1].transformer_list[0][1]
        clf = pipe.steps[-1][1]

        # cmd line args have precedence over defaults
        assert scaler.copy == 555
        assert clf.fit_intercept == 456
        assert clf.normalize == 789



"""Tests for callbacks in training.py"""

from functools import partial
import pickle
from unittest.mock import Mock
from unittest.mock import patch
from unittest.mock import call
from copy import deepcopy

import numpy as np
import pytest
from sklearn.base import clone
from sklearn.model_selection import train_test_split
from sklearn.metrics import log_loss
import torch
from torch.utils.data import TensorDataset

from skorch.helper import predefined_split


class TestCheckpoint:
    @pytest.fixture
    def checkpoint_cls(self):
        from skorch.callbacks import Checkpoint
        return Checkpoint

    @pytest.fixture
    def save_params_mock(self):
        with patch('skorch.NeuralNet.save_params') as mock:
            yield mock

    @pytest.fixture
    def pickle_dump_mock(self):
        with patch('pickle.dump') as mock:
            yield mock

    @pytest.fixture
    def net_cls(self):
        """very simple network that trains for 10 epochs"""
        from skorch import NeuralNetRegressor
        from skorch.toy import make_regressor

        module_cls = make_regressor(
            input_units=1,
            num_hidden=0,
            output_units=1,
        )

        return partial(
            NeuralNetRegressor,
            module=module_cls,
            max_epochs=10,
            batch_size=10)

    @pytest.fixture(scope='module')
    def data(self):
        # have 10 examples so we can do a nice CV split
        X = np.zeros((10, 1), dtype='float32')
        y = np.zeros((10, 1), dtype='float32')
        return X, y

    def test_init_with_wrong_kwarg_name_raises(self, checkpoint_cls):
        checkpoint_cls(f_foobar='foobar.pt').initialize()  # works
        msg = "Checkpoint got an unexpected argument 'foobar', did you mean 'f_foobar'?"
        with pytest.raises(TypeError, match=msg):
            checkpoint_cls(foobar='foobar.pt').initialize()

    def test_init_with_f_params_and_f_module_raises(self, checkpoint_cls):
        msg = "Checkpoint called with both f_params and f_module, please choose one"
        with pytest.raises(TypeError, match=msg):
            checkpoint_cls(f_module='weights.pt', f_params='params.pt').initialize()

    def test_none_monitor_saves_always(
            self, save_params_mock, net_cls, checkpoint_cls, data):
        sink = Mock()
        net = net_cls(callbacks=[
            checkpoint_cls(monitor=None, sink=sink,
                           event_name='event_another'),
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 4 * len(net.history)
        assert sink.call_count == len(net.history)
        assert all((x is True) for x in net.history[:, 'event_another'])

    @pytest.mark.parametrize('message,files', [
        ('Unable to save module state to params.pt, '
         'Exception: encoding error',
         {
             'f_params': 'params.pt',
             'f_optimizer': None,
             'f_criterion': None,
             'f_history': None,
         }),
        ('Unable to save optimizer state to optimizer.pt, '
         'Exception: encoding error',
         {
             'f_params': None,
             'f_optimizer': 'optimizer.pt',
             'f_criterion': None,
             'f_history': None,
         }),
        ('Unable to save criterion state to criterion.pt, '
         'Exception: encoding error',
         {
             'f_params': None,
             'f_optimizer': None,
             'f_criterion': 'criterion.pt',
             'f_history': None,
         }),
        ('Unable to save history to history.json, '
         'Exception: encoding error',
         {
             'f_params': None,
             'f_optimizer': None,
             'f_criterion': None,
             'f_history': 'history.json',
         }),
    ])
    def test_outputs_to_sink_when_save_params_errors(
            self, save_params_mock, net_cls, checkpoint_cls, data,
            message, files):
        sink = Mock()
        save_params_mock.side_effect = Exception('encoding error')
        net = net_cls(callbacks=[
            checkpoint_cls(monitor=None, sink=sink, **files)
        ])
        net.fit(*data)

        assert save_params_mock.call_count == len(net.history)
        assert sink.call_count == 2*len(net.history)
        save_error_messages = [call(message)] * len(net.history)
        sink.assert_has_calls(save_error_messages, any_order=True)

    @pytest.mark.parametrize('f_name, mode', [
        ('f_params', 'w'),
        ('f_optimizer', 'w'),
        ('f_criterion', 'w'),
        ('f_history', 'w'),
        ('f_pickle', 'wb')
    ])
    def test_init_with_dirname_and_file_like_object_error(
            self, checkpoint_cls, tmpdir, f_name, mode):
        from skorch.exceptions import SkorchException

        skorch_dir = tmpdir.mkdir("skorch")
        exp_dir = skorch_dir.join("exp1")
        f = skorch_dir.join(f_name + ".pt")

        with f.open(mode) as fp:
            with pytest.raises(SkorchException) as e:
                checkpoint_cls(**{f_name: fp}, dirname=str(exp_dir))
        expected = "dirname can only be used when f_* are strings"
        assert str(e.value) == expected

    @pytest.mark.parametrize('f_name, mode', [
        ('f_params', 'w'),
        ('f_optimizer', 'w'),
        ('f_criterion', 'w'),
        ('f_history', 'w'),
        ('f_pickle', 'wb')
    ])
    def test_initialize_with_dirname_and_file_like_object_error(
            self, checkpoint_cls, tmpdir, f_name, mode):
        from skorch.exceptions import SkorchException

        skorch_dir = tmpdir.mkdir("skorch")
        exp_dir = skorch_dir.join("exp1")
        f = skorch_dir.join(f_name + ".pt")

        with f.open(mode) as fp:
            with pytest.raises(SkorchException) as e:
                cp = checkpoint_cls(dirname=str(exp_dir))
                setattr(cp, f_name, fp)
                cp.initialize()
        expected = "dirname can only be used when f_* are strings"
        assert str(e.value) == expected

    def test_default_without_validation_raises_meaningful_error(
            self, net_cls, checkpoint_cls, data):
        net = net_cls(
            callbacks=[
                checkpoint_cls(),
            ],
            train_split=None
        )
        from skorch.exceptions import SkorchException

        msg_expected = (
            r"Key 'valid_loss_best' was not found in history. "
            r"Make sure you have validation data if you use "
            r"validation scores for checkpointing."
        )
        with pytest.raises(SkorchException, match=msg_expected):
            net.fit(*data)

    def test_string_monitor_and_formatting(
            self, save_params_mock, net_cls, checkpoint_cls, data):
        def epoch_3_scorer(net, *_):
            return 1 if net.history[-1, 'epoch'] == 3 else 0

        from skorch.callbacks import EpochScoring
        scoring = EpochScoring(
            scoring=epoch_3_scorer, on_train=True, lower_is_better=False)

        sink = Mock()
        cb = checkpoint_cls(
            monitor='epoch_3_scorer_best',
            f_params='model_{last_epoch[epoch]}_{net.max_epochs}.pt',
            f_optimizer='optimizer_{last_epoch[epoch]}_{net.max_epochs}.pt',
            f_criterion='criterion_{last_epoch[epoch]}_{net.max_epochs}.pt',
            sink=sink)
        net = net_cls(callbacks=[
            ('my_score', scoring), cb
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 8
        assert cb.get_formatted_files(net) == {
            'f_params': 'model_3_10.pt',
            'f_optimizer': 'optimizer_3_10.pt',
            'f_criterion': 'criterion_3_10.pt',
            'f_history': 'history.json',
            'f_pickle': None
        }
        save_params_mock.assert_has_calls(
            [
                call(f_module='model_1_10.pt'),  # params is turned into module
                call(f_optimizer='optimizer_1_10.pt'),
                call(f_criterion='criterion_1_10.pt'),
                call(f_history='history.json'),
                call(f_module='model_3_10.pt'),  # params is turned into module
                call(f_optimizer='optimizer_3_10.pt'),
                call(f_criterion='criterion_3_10.pt'),
                call(f_history='history.json'),
            ],
            any_order=True,
        )
        assert sink.call_count == 2
        # The first epoch will always be saved. `epoch_3_scorer` returns 1 at
        # epoch 3, which will trigger another checkpoint. For all other epochs
        # `epoch_3_scorer` returns 0, which does not trigger a checkpoint.
        assert [True, False, True] + [False] * 7 == net.history[:, 'event_cp']

    def test_save_all_targets(
            self, save_params_mock, pickle_dump_mock,
            net_cls, checkpoint_cls, data):
        net = net_cls(callbacks=[
            checkpoint_cls(
                monitor=None,
                f_params='params.pt',
                f_pickle='model.pkl',
                f_optimizer='optimizer.pt',
                f_criterion='criterion.pt',
                f_history='history.json',
            ),
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 4 * len(net.history)
        assert pickle_dump_mock.call_count == len(net.history)

        save_params_mock.assert_has_calls(
            [
                call(f_module='params.pt'),  # params is turned into module
                call(f_optimizer='optimizer.pt'),
                call(f_criterion='criterion.pt'),
                call(f_history='history.json'),
            ] * len(net.history),
            any_order=True,
        )

    def test_save_all_targets_with_prefix(
            self, save_params_mock, pickle_dump_mock,
            net_cls, checkpoint_cls, data):

        cp = checkpoint_cls(
            monitor=None,
            f_params='params.pt',
            f_pickle='model.pkl',
            f_optimizer='optimizer.pt',
            f_criterion='criterion.pt',
            f_history='history.json',
            fn_prefix="exp1_",
        )
        net = net_cls(callbacks=[cp])
        net.fit(*data)

        assert cp.f_history_ == "exp1_history.json"
        assert save_params_mock.call_count == 4 * len(net.history)
        assert pickle_dump_mock.call_count == len(net.history)
        save_params_mock.assert_has_calls(
            [
                call(f_module='exp1_params.pt'),  # params is turned into module
                call(f_optimizer='exp1_optimizer.pt'),
                call(f_criterion='exp1_criterion.pt'),
                call(f_history='exp1_history.json'),
            ] * len(net.history),
            any_order=True,
        )

    def test_save_all_targets_with_prefix_and_dirname(
            self, save_params_mock, pickle_dump_mock,
            net_cls, checkpoint_cls, data, tmpdir):

        skorch_dir = tmpdir.mkdir('skorch').join('exp1')

        cp = checkpoint_cls(
            monitor=None,
            f_params='params.pt',
            f_history='history.json',
            f_pickle='model.pkl',
            f_optimizer='optimizer.pt',
            f_criterion='criterion.pt',
            fn_prefix="unet_",
            dirname=str(skorch_dir))
        net = net_cls(callbacks=[cp])
        net.fit(*data)

        f_params = skorch_dir.join('unet_params.pt')
        f_optimizer = skorch_dir.join('unet_optimizer.pt')
        f_criterion = skorch_dir.join('unet_criterion.pt')
        f_history = skorch_dir.join('unet_history.json')

        assert cp.f_history_ == str(f_history)
        assert save_params_mock.call_count == 4 * len(net.history)
        assert pickle_dump_mock.call_count == len(net.history)
        save_params_mock.assert_has_calls(
            [
                call(f_module=str(f_params)),  # params is turned into module
                call(f_optimizer=str(f_optimizer)),
                call(f_criterion=str(f_criterion)),
                call(f_history=str(f_history)),
            ] * len(net.history),
            any_order=True,
        )
        assert skorch_dir.exists()

    def test_save_no_targets(
            self, save_params_mock, pickle_dump_mock,
            net_cls, checkpoint_cls, data):
        net = net_cls(callbacks=[
            checkpoint_cls(
                monitor=None,
                f_params=None,
                f_optimizer=None,
                f_criterion=None,
                f_history=None,
                f_pickle=None,
            ),
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 0
        assert pickle_dump_mock.call_count == 0

    def test_warnings_when_monitor_appears_in_history(
            self, net_cls, checkpoint_cls, save_params_mock, data):
        net = net_cls(
            callbacks=[checkpoint_cls(monitor="valid_loss")],
            max_epochs=1)

        exp_warn = (
            "Checkpoint monitor parameter is set to 'valid_loss' and the "
            "history contains 'valid_loss_best'. Perhaps you meant to set the "
            "parameter to 'valid_loss_best'")

        with pytest.warns(UserWarning, match=exp_warn):
            net.fit(*data)
        assert save_params_mock.call_count == 4

    def test_save_custom_module(
            self, save_params_mock, module_cls, checkpoint_cls, data,
    ):
        # checkpointing custom modules works
        from skorch import NeuralNetRegressor

        class MyNet(NeuralNetRegressor):
            """Net with custom module"""
            def __init__(self, *args, mymodule=module_cls, **kwargs):
                self.mymodule = mymodule
                super().__init__(*args, **kwargs)

            def initialize_module(self, *args, **kwargs):
                super().initialize_module(*args, **kwargs)

                params = self.get_params_for('mymodule')
                self.mymodule_ = self.mymodule(**params)

                return self

        cp = checkpoint_cls(
            monitor=None,
            f_params=None,
            f_optimizer=None,
            f_criterion=None,
            f_history=None,
            f_mymodule='mymodule.pt',
        )
        net = MyNet(module_cls, callbacks=[cp])
        net.fit(*data)

        assert save_params_mock.call_count == 1 * len(net.history)
        save_params_mock.assert_has_calls(
            [call(f_mymodule='mymodule.pt')] * len(net.history)
        )

    @pytest.fixture
    def load_params(self):
        import torch
        return torch.load

    @pytest.mark.parametrize('load_best_flag', [False, True])
    def test_automatically_load_checkpoint(
            self, net_cls, checkpoint_cls, data, tmp_path,
            load_params, load_best_flag,
    ):
        # checkpoint once at the beginning of training.
        # when restoring at the end of training, the parameters
        # of the net should not differ. If we do not restore
        # then the parameters must differ.
        path_cb = tmp_path / 'params_cb.pt'
        path_net = tmp_path / 'params_net.pt'

        def save_once_monitor(net):
            return len(net.history) == 1

        net = net_cls(
            max_epochs=3,
            callbacks=[
                checkpoint_cls(
                    monitor=save_once_monitor,
                    f_params=path_cb,
                    load_best=load_best_flag,
                ),
            ],
        )

        net.fit(*data)
        net.save_params(path_net)

        params_cb = load_params(path_cb)
        params_net = load_params(path_net)

        if load_best_flag:
            assert params_cb == params_net
        else:
            assert params_cb != params_net



class TestEarlyStopping:

    @pytest.fixture
    def early_stopping_cls(self):
        from skorch.callbacks import EarlyStopping
        return EarlyStopping

    @pytest.fixture
    def epoch_scoring_cls(self):
        from skorch.callbacks import EpochScoring
        return EpochScoring

    @pytest.fixture
    def net_clf_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def broken_classifier_module(self, classifier_module):
        """Return a classifier that does not improve over time."""
        class BrokenClassifier(classifier_module.func):
            def forward(self, x):
                return super().forward(x) * 0 + 0.5
        return BrokenClassifier

    def test_typical_use_case_nonstop(
            self, net_clf_cls, classifier_module, classifier_data,
            early_stopping_cls):
        patience = 5
        max_epochs = 8
        early_stopping_cb = early_stopping_cls(patience=patience)

        net = net_clf_cls(
            classifier_module,
            callbacks=[
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )
        net.fit(*classifier_data)

        assert len(net.history) == max_epochs

    def test_weights_restore(
            self, net_clf_cls, classifier_module, classifier_data,
            early_stopping_cls):
        patience = 3
        max_epochs = 20
        seed = 1

        side_effect = []

        def sink(x):
            side_effect.append(x)

        early_stopping_cb = early_stopping_cls(
            patience=patience,
            sink=sink,
            load_best=True,
            monitor="valid_acc",
            lower_is_better=False,
        )

        # Split dataset to have a fixed validation
        X_tr, X_val, y_tr, y_val = train_test_split(
            *classifier_data, random_state=seed)
        tr_dataset = TensorDataset(
            torch.as_tensor(X_tr).float(), torch.as_tensor(y_tr))
        val_dataset = TensorDataset(
            torch.as_tensor(X_val).float(), torch.as_tensor(y_val))

        # Fix the network once with early stoppping and fixed seed
        net1 = net_clf_cls(
            classifier_module,
            callbacks=[early_stopping_cb],
            max_epochs=max_epochs,
            train_split=predefined_split(val_dataset),
        )
        torch.manual_seed(seed)
        net1.fit(tr_dataset, y=None)

        # Check training was stopped before the end
        assert len(net1.history) < max_epochs

        # check correct output messages
        assert len(side_effect) == 2

        msg = side_effect[0]
        expected_msg = ("Stopping since valid_acc has not improved in "
                        "the last 3 epochs.")
        assert msg == expected_msg

        msg = side_effect[1]
        expected_msg = "Restoring best model from epoch "
        assert expected_msg in msg

        # Recompute validation loss and store it together with module weights
        y_proba = net1.predict_proba(val_dataset)
        es_weights = deepcopy(net1.module_.state_dict())
        es_loss = log_loss(y_val, y_proba)

        # Retrain same classifier without ES, using the best epochs number
        net2 = net_clf_cls(
            classifier_module,
            max_epochs=early_stopping_cb.best_epoch_,
            train_split=predefined_split(val_dataset),
        )
        torch.manual_seed(seed)
        net2.fit(tr_dataset, y=None)

        # Check that weights obtained match
        assert all(
            torch.equal(wi, wj)
            for wi, wj in zip(
                net2.module_.state_dict().values(),
                es_weights.values()
            )
        )

        # Check validation loss obtained match
        y_proba_2 = net2.predict_proba(val_dataset)
        assert es_loss == log_loss(y_val, y_proba_2)

        # Check best_model_weights_ is transformed into None when pickling
        del net1.callbacks[0].sink
        net1_pkl = pickle.dumps(net1)

        reloaded_net1 = pickle.loads(net1_pkl)
        assert reloaded_net1.callbacks[0].best_epoch_ == net1.callbacks[0].best_epoch_
        assert reloaded_net1.callbacks[0].best_model_weights_ is None

    def test_typical_use_case_stopping(
            self, net_clf_cls, broken_classifier_module, classifier_data,
            early_stopping_cls):
        patience = 5
        max_epochs = 8
        side_effect = []

        def sink(x):
            side_effect.append(x)

        early_stopping_cb = early_stopping_cls(patience=patience, sink=sink)

        net = net_clf_cls(
            broken_classifier_module,
            callbacks=[
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )
        net.fit(*classifier_data)

        assert len(net.history) == patience + 1 < max_epochs

        # check correct output message
        assert len(side_effect) == 1
        msg = side_effect[0]
        expected_msg = ("Stopping since valid_loss has not improved in "
                        "the last 5 epochs.")
        assert msg == expected_msg

    def test_custom_scoring_nonstop(
            self, net_clf_cls, classifier_module, classifier_data,
            early_stopping_cls, epoch_scoring_cls,
    ):
        lower_is_better = False
        scoring_name = 'valid_roc_auc'
        patience = 5
        max_epochs = 8
        scoring_mock = Mock(side_effect=list(range(2, 10)))
        scoring_cb = epoch_scoring_cls(
            scoring_mock, lower_is_better, name=scoring_name)
        early_stopping_cb = early_stopping_cls(
            patience=patience, lower_is_better=lower_is_better,
            monitor=scoring_name)

        net = net_clf_cls(
            classifier_module,
            callbacks=[
                scoring_cb,
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )
        net.fit(*classifier_data)

        assert len(net.history) == max_epochs

    def test_custom_scoring_stop(
            self, net_clf_cls, broken_classifier_module, classifier_data,
            early_stopping_cls, epoch_scoring_cls,
    ):
        lower_is_better = False
        scoring_name = 'valid_roc_auc'
        patience = 5
        max_epochs = 8
        scoring_cb = epoch_scoring_cls(
            'roc_auc', lower_is_better, name=scoring_name)
        early_stopping_cb = early_stopping_cls(
            patience=patience, lower_is_better=lower_is_better,
            monitor=scoring_name)

        net = net_clf_cls(
            broken_classifier_module,
            callbacks=[
                scoring_cb,
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )
        net.fit(*classifier_data)

        assert len(net.history) < max_epochs

    def test_stopping_big_absolute_threshold(
            self, net_clf_cls, classifier_module, classifier_data,
            early_stopping_cls):
        patience = 5
        max_epochs = 8
        early_stopping_cb = early_stopping_cls(patience=patience,
                                               threshold_mode='abs',
                                               threshold=0.1)

        net = net_clf_cls(
            classifier_module,
            callbacks=[
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )
        net.fit(*classifier_data)

        assert len(net.history) == patience + 1 < max_epochs

    def test_wrong_threshold_mode(
            self, net_clf_cls, classifier_module, classifier_data,
            early_stopping_cls):
        patience = 5
        max_epochs = 8
        early_stopping_cb = early_stopping_cls(
            patience=patience, threshold_mode='incorrect')
        net = net_clf_cls(
            classifier_module,
            callbacks=[
                early_stopping_cb,
            ],
            max_epochs=max_epochs,
        )

        with pytest.raises(ValueError) as exc:
            net.fit(*classifier_data)

        expected_msg = "Invalid threshold mode: 'incorrect'"
        assert exc.value.args[0] == expected_msg



class TestParamMapper:

    @pytest.fixture
    def initializer(self):
        from skorch.callbacks import Initializer
        return Initializer

    @pytest.fixture
    def freezer(self):
        from skorch.callbacks import Freezer
        return Freezer

    @pytest.fixture
    def unfreezer(self):
        from skorch.callbacks import Unfreezer
        return Unfreezer

    @pytest.fixture
    def param_mapper(self):
        from skorch.callbacks import ParamMapper
        return ParamMapper

    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.mark.parametrize('at', [0, -1])
    def test_subzero_at_fails(self, net_cls, classifier_module,
                              param_mapper, at):
        cb = param_mapper(patterns='*', at=at)
        net = net_cls(classifier_module, callbacks=[cb])
        with pytest.raises(ValueError):
            net.initialize()

    @pytest.mark.parametrize('mod_init', [False, True])
    @pytest.mark.parametrize('weight_pattern', [
        'sequential.*.weight',
        lambda name: name.startswith('sequential') and name.endswith('.weight'),
    ])
    def test_initialization_is_effective(self, net_cls, classifier_module,
                                         classifier_data, initializer,
                                         mod_init, weight_pattern):
        from torch.nn.init import constant_
        from skorch.utils import to_numpy

        module = classifier_module() if mod_init else classifier_module

        net = net_cls(
            module,
            lr=0,
            max_epochs=1,
            callbacks=[
                initializer(weight_pattern, partial(constant_, val=5)),
                initializer('sequential.3.bias', partial(constant_, val=10)),
            ])

        net.fit(*classifier_data)

        assert np.allclose(to_numpy(net.module_.sequential[0].weight), 5)
        assert np.allclose(to_numpy(net.module_.sequential[3].weight), 5)
        assert np.allclose(to_numpy(net.module_.sequential[3].bias), 10)

    @pytest.mark.parametrize('mod_init', [False, True])
    @pytest.mark.parametrize('mod_kwargs', [
        {},
        # Supply a module__ parameter so the model is forced
        # to re-initialize. Even then parameters should be
        # frozen correctly.
        {'module__hidden_units': 5},
    ])
    def test_freezing_is_effective(self, net_cls, classifier_module,
                                   classifier_data, freezer, mod_init,
                                   mod_kwargs):
        from skorch.utils import to_numpy

        module = classifier_module() if mod_init else classifier_module

        net = net_cls(
            module,
            max_epochs=2,
            callbacks=[
                freezer('sequential.*.weight'),
                freezer('sequential.3.bias'),
            ],
            **mod_kwargs)

        net.initialize()

        assert net.module_.sequential[0].weight.requires_grad
        assert net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert net.module_.sequential[3].bias.requires_grad

        dense0_weight_pre = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_pre = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_pre = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_pre = to_numpy(net.module_.sequential[3].bias).copy()

        # use partial_fit to not re-initialize the module (weights)
        net.partial_fit(*classifier_data)

        dense0_weight_post = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_post = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_post = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_post = to_numpy(net.module_.sequential[3].bias).copy()

        assert not net.module_.sequential[0].weight.requires_grad
        assert not net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert not net.module_.sequential[3].bias.requires_grad

        assert np.allclose(dense0_weight_pre, dense0_weight_post)
        assert np.allclose(dense1_weight_pre, dense1_weight_post)
        assert not np.allclose(dense0_bias_pre, dense0_bias_post)
        assert np.allclose(dense1_bias_pre, dense1_bias_post)

    def test_unfreezing_is_effective(self, net_cls, classifier_module,
                                     classifier_data, freezer, unfreezer):
        from skorch.utils import to_numpy

        net = net_cls(
            classifier_module,
            max_epochs=1,
            callbacks=[
                freezer('sequential.*.weight'),
                freezer('sequential.3.bias'),
                unfreezer('sequential.*.weight', at=2),
                unfreezer('sequential.3.bias', at=2),
            ])

        net.initialize()

        # epoch 1, freezing parameters
        net.partial_fit(*classifier_data)

        assert not net.module_.sequential[0].weight.requires_grad
        assert not net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert not net.module_.sequential[3].bias.requires_grad

        dense0_weight_pre = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_pre = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_pre = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_pre = to_numpy(net.module_.sequential[3].bias).copy()

        # epoch 2, unfreezing parameters
        net.partial_fit(*classifier_data)

        assert net.module_.sequential[0].weight.requires_grad
        assert net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert net.module_.sequential[3].bias.requires_grad

        # epoch 3, modifications should have been made
        net.partial_fit(*classifier_data)

        dense0_weight_post = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_post = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_post = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_post = to_numpy(net.module_.sequential[3].bias).copy()

        assert not np.allclose(dense0_weight_pre, dense0_weight_post)
        assert not np.allclose(dense1_weight_pre, dense1_weight_post)
        assert not np.allclose(dense0_bias_pre, dense0_bias_post)
        assert not np.allclose(dense1_bias_pre, dense1_bias_post)


    def test_schedule_is_effective(self, net_cls, classifier_module,
                                   classifier_data, param_mapper):
        from skorch.utils import to_numpy, noop
        from skorch.utils import freeze_parameter, unfreeze_parameter

        def schedule(net):
            if len(net.history) == 1:
                return freeze_parameter
            elif len(net.history) == 2:
                return unfreeze_parameter
            return noop

        net = net_cls(
            classifier_module,
            max_epochs=1,
            callbacks=[
                param_mapper(
                    ['sequential.*.weight', 'sequential.3.bias'],
                    schedule=schedule,
                ),
            ])

        net.initialize()

        # epoch 1, freezing parameters
        net.partial_fit(*classifier_data)

        assert not net.module_.sequential[0].weight.requires_grad
        assert not net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert not net.module_.sequential[3].bias.requires_grad

        dense0_weight_pre = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_pre = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_pre = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_pre = to_numpy(net.module_.sequential[3].bias).copy()

        # epoch 2, unfreezing parameters
        net.partial_fit(*classifier_data)

        assert net.module_.sequential[0].weight.requires_grad
        assert net.module_.sequential[3].weight.requires_grad
        assert net.module_.sequential[0].bias.requires_grad
        assert net.module_.sequential[3].bias.requires_grad

        # epoch 3, modifications should have been made
        net.partial_fit(*classifier_data)

        dense0_weight_post = to_numpy(net.module_.sequential[0].weight).copy()
        dense1_weight_post = to_numpy(net.module_.sequential[3].weight).copy()
        dense0_bias_post = to_numpy(net.module_.sequential[0].bias).copy()
        dense1_bias_post = to_numpy(net.module_.sequential[3].bias).copy()

        assert not np.allclose(dense0_weight_pre, dense0_weight_post)
        assert not np.allclose(dense1_weight_pre, dense1_weight_post)
        assert not np.allclose(dense0_bias_pre, dense0_bias_post)
        assert not np.allclose(dense1_bias_pre, dense1_bias_post)


class TestLoadInitState:

    @pytest.fixture
    def checkpoint_cls(self):
        from skorch.callbacks import Checkpoint
        return Checkpoint

    @pytest.fixture
    def loadinitstate_cls(self):
        from skorch.callbacks import LoadInitState
        return LoadInitState

    @pytest.fixture
    def net_cls(self):
        """very simple network that trains for 10 epochs"""
        from skorch import NeuralNetRegressor
        from skorch.toy import make_regressor

        module_cls = make_regressor(
            input_units=1,
            num_hidden=0,
            output_units=1,
        )

        return partial(
            NeuralNetRegressor,
            module=module_cls,
            max_epochs=10,
            batch_size=10)

    @pytest.fixture(scope='module')
    def data(self):
        # have 10 examples so we can do a nice CV split
        X = np.zeros((10, 1), dtype='float32')
        y = np.zeros((10, 1), dtype='float32')
        return X, y

    def test_load_initial_state(
            self, checkpoint_cls, net_cls, loadinitstate_cls,
            data, tmpdir):
        skorch_dir = tmpdir.mkdir('skorch')
        f_params = skorch_dir.join('params.pt')
        f_optimizer = skorch_dir.join('optimizer.pt')
        f_criterion = skorch_dir.join('criterion.pt')
        f_history = skorch_dir.join('history.json')

        cp = checkpoint_cls(
            monitor=None,
            f_params=str(f_params),
            f_optimizer=str(f_optimizer),
            f_criterion=str(f_criterion),
            f_history=str(f_history)
        )
        load_init_state = loadinitstate_cls(cp)
        net = net_cls(callbacks=[cp, load_init_state])
        net.fit(*data)

        assert f_params.exists()
        assert f_optimizer.exists()
        assert f_criterion.exists()
        assert f_history.exists()

        assert len(net.history) == 10
        del net

        new_net = net_cls(callbacks=[cp, load_init_state])
        new_net.fit(*data)

        assert len(new_net.history) == 20

    def test_load_initial_state_custom_scoring(
            self, checkpoint_cls, net_cls, loadinitstate_cls,
            data, tmpdir):
        def epoch_3_scorer(net, *_):
            return 1 if net.history[-1, 'epoch'] == 3 else 0

        from skorch.callbacks import EpochScoring
        scoring = EpochScoring(
            scoring=epoch_3_scorer, on_train=True, lower_is_better=False)

        skorch_dir = tmpdir.mkdir('skorch')
        f_params = skorch_dir.join(
            'model_epoch_{last_epoch[epoch]}.pt')
        f_optimizer = skorch_dir.join(
            'optimizer_epoch_{last_epoch[epoch]}.pt')
        f_criterion = skorch_dir.join(
            'criterion_epoch_{last_epoch[epoch]}.pt')
        f_history = skorch_dir.join(
            'history.json')

        cp = checkpoint_cls(
            monitor='epoch_3_scorer_best',
            f_params=str(f_params),
            f_optimizer=str(f_optimizer),
            f_criterion=str(f_criterion),
            f_history=str(f_history)
        )
        load_init_state = loadinitstate_cls(cp)
        net = net_cls(callbacks=[load_init_state, scoring, cp])

        net.fit(*data)

        assert skorch_dir.join('model_epoch_3.pt').exists()
        assert skorch_dir.join('optimizer_epoch_3.pt').exists()
        assert skorch_dir.join('criterion_epoch_3.pt').exists()
        assert skorch_dir.join('history.json').exists()

        assert len(net.history) == 10
        del net

        new_net = net_cls(callbacks=[load_init_state, scoring, cp])
        new_net.fit(*data)

        # new_net starts from the best epoch of the first run
        # the best epcoh of the previous run was at epoch 3
        # the second run went through 10 epochs, thus
        # 3 + 10 = 13
        assert len(new_net.history) == 13
        assert new_net.history[:, 'event_cp'] == [
            True, False, True] + [False] * 10


class TestTrainEndCheckpoint:
    @pytest.fixture
    def trainendcheckpoint_cls(self):
        from skorch.callbacks import TrainEndCheckpoint
        return TrainEndCheckpoint

    @pytest.fixture
    def save_params_mock(self):
        with patch('skorch.NeuralNet.save_params') as mock:
            yield mock

    @pytest.fixture
    def pickle_dump_mock(self):
        with patch('pickle.dump') as mock:
            yield mock

    @pytest.fixture
    def net_cls(self):
        """very simple network that trains for 10 epochs"""
        from skorch import NeuralNetRegressor
        from skorch.toy import make_regressor

        module_cls = make_regressor(
            input_units=1,
            num_hidden=0,
            output_units=1,
        )

        return partial(
            NeuralNetRegressor,
            module=module_cls,
            max_epochs=10,
            batch_size=10)

    @pytest.fixture(scope='module')
    def data(self):
        # have 10 examples so we can do a nice CV split
        X = np.zeros((10, 1), dtype='float32')
        y = np.zeros((10, 1), dtype='float32')
        return X, y

    def test_init_with_wrong_kwarg_name_raises(self, trainendcheckpoint_cls):
        trainendcheckpoint_cls(f_foobar='foobar.pt').initialize()  # works
        msg = ("TrainEndCheckpoint got an unexpected argument 'foobar', "
               "did you mean 'f_foobar'?")
        with pytest.raises(TypeError, match=msg):
            trainendcheckpoint_cls(foobar='foobar.pt').initialize()

    def test_init_with_f_params_and_f_module_raises(self, trainendcheckpoint_cls):
        msg = "Checkpoint called with both f_params and f_module, please choose one"
        with pytest.raises(TypeError, match=msg):
            trainendcheckpoint_cls(
                f_module='weights.pt', f_params='params.pt').initialize()

    def test_saves_at_end(
            self, save_params_mock, net_cls, trainendcheckpoint_cls, data):
        sink = Mock()
        net = net_cls(callbacks=[
            trainendcheckpoint_cls(
                sink=sink, dirname='exp1', fn_prefix='train_end_')
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 4
        assert sink.call_args == call("Final checkpoint triggered")
        save_params_mock.assert_has_calls(
            [
                # params is turned into module
                call(f_module='exp1/train_end_params.pt'),
                call(f_optimizer='exp1/train_end_optimizer.pt'),
                call(f_criterion='exp1/train_end_criterion.pt'),
                call(f_history='exp1/train_end_history.json'),
            ],
            any_order=True,
        )

    def test_saves_at_end_with_custom_formatting(
            self, save_params_mock, net_cls, trainendcheckpoint_cls, data):
        sink = Mock()
        net = net_cls(callbacks=[
            trainendcheckpoint_cls(
                sink=sink,
                dirname='exp1',
                f_params='model_{last_epoch[epoch]}.pt',
                f_optimizer='optimizer_{last_epoch[epoch]}.pt',
                f_criterion='criterion_{last_epoch[epoch]}.pt',
                fn_prefix='train_end_',
            )
        ])
        net.fit(*data)

        assert save_params_mock.call_count == 4
        assert sink.call_args == call("Final checkpoint triggered")
        save_params_mock.assert_has_calls(
            [
                # params is turned into module
                call(f_module='exp1/train_end_model_10.pt'),
                call(f_optimizer='exp1/train_end_optimizer_10.pt'),
                call(f_criterion='exp1/train_end_criterion_10.pt'),
                call(f_history='exp1/train_end_history.json'),
            ],
            any_order=True,
        )

    def test_cloneable(self, trainendcheckpoint_cls):
        # reproduces bug #459
        cp = trainendcheckpoint_cls()
        clone(cp)  # does not raise

    def test_train_end_with_load_init(self, trainendcheckpoint_cls, net_cls, data):
        # test for https://github.com/skorch-dev/skorch/issues/528
        # Check that the initial state is indeed loaded from the checkpoint.
        from skorch.callbacks import LoadInitState
        from sklearn.metrics import mean_squared_error

        X, y = data
        cp = trainendcheckpoint_cls()
        net = net_cls(callbacks=[cp], max_epochs=3, lr=0.1).initialize()
        score_before = mean_squared_error(y, net.predict(X))
        net.partial_fit(X, y)
        score_after = mean_squared_error(y, net.predict(X))

        # make sure the net learned at all
        assert score_after < score_before

        net_new = net_cls(callbacks=[LoadInitState(cp)], max_epochs=0)
        net_new.fit(X, y)
        score_loaded = mean_squared_error(y, net_new.predict(X))

        # the same score as after the end of training of the initial
        # net should be obtained
        assert np.isclose(score_loaded, score_after)

    def test_save_custom_module(
            self, save_params_mock, module_cls, trainendcheckpoint_cls, data,
    ):
        # checkpointing custom modules works
        from skorch import NeuralNetRegressor

        class MyNet(NeuralNetRegressor):
            """Net with custom module"""
            def __init__(self, *args, mymodule=module_cls, **kwargs):
                self.mymodule = mymodule
                super().__init__(*args, **kwargs)

            def initialize_module(self, *args, **kwargs):
                super().initialize_module(*args, **kwargs)

                params = self.get_params_for('mymodule')
                self.mymodule_ = self.mymodule(**params)

                return self

        cp = trainendcheckpoint_cls(
            f_params=None,
            f_optimizer=None,
            f_criterion=None,
            f_history=None,
            f_mymodule='mymodule.pt',
        )
        net = MyNet(module_cls, callbacks=[cp])
        net.fit(*data)

        assert save_params_mock.call_count == 1
        save_params_mock.assert_has_calls([call(f_mymodule='train_end_mymodule.pt')])

    def test_pickle_uninitialized_callback(self, trainendcheckpoint_cls):
        # isuue 773
        cp = trainendcheckpoint_cls()
        # does not raise
        s = pickle.dumps(cp)
        pickle.loads(s)

    def test_pickle_initialized_callback(self, trainendcheckpoint_cls):
        # issue 773
        cp = trainendcheckpoint_cls().initialize()
        # does not raise
        s = pickle.dumps(cp)
        pickle.loads(s)


class TestInputShapeSetter:

    @pytest.fixture
    def module_cls(self):
        import torch

        class Module(torch.nn.Module):
            def __init__(self, input_dim=3):
                super().__init__()
                self.layer = torch.nn.Linear(input_dim, 2)
            def forward(self, X):
                return self.layer(X)

        return Module

    @pytest.fixture
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture
    def input_shape_setter_cls(self):
        from skorch.callbacks import InputShapeSetter
        return InputShapeSetter

    def generate_data(self, n_input):
        from sklearn.datasets import make_classification
        X, y = make_classification(
            1000,
            n_input,
            n_informative=n_input,
            n_redundant=0,
            random_state=0,
        )
        return X.astype(np.float32), y

    @pytest.fixture
    def data_fixed(self):
        return self.generate_data(n_input=10)

    @pytest.fixture(params=[2, 10, 20])
    def data_parametrized(self, request):
        return self.generate_data(n_input=request.param)

    def test_shape_set(
        self, net_cls, module_cls, input_shape_setter_cls, data_parametrized,
    ):
        net = net_cls(module_cls, max_epochs=2, callbacks=[
            input_shape_setter_cls(),
        ])

        X, y = data_parametrized
        n_input = X.shape[1]
        net.fit(X, y)

        assert net.module_.layer.in_features == n_input

    def test_one_dimensional_x_raises(
        self, net_cls, module_cls, input_shape_setter_cls,
    ):
        net = net_cls(module_cls, max_epochs=2, callbacks=[
            input_shape_setter_cls(),
        ])

        X, y = np.zeros(10), np.zeros(10)

        with pytest.raises(ValueError) as e:
            net.fit(X, y)

        assert (
            "Expected at least two-dimensional input data for X. "
            "If your data is one-dimensional, please use the `input_dim_fn` "
            "parameter to infer the correct input shape."
            ) in str(e)

    def test_shape_set_using_fn(
        self, net_cls, module_cls, input_shape_setter_cls, data_parametrized,
    ):
        fn_calls = 0

        def input_dim_fn(X):
            nonlocal fn_calls
            fn_calls += 1
            return X.shape[1]

        net = net_cls(module_cls, max_epochs=2, callbacks=[
            input_shape_setter_cls(input_dim_fn=input_dim_fn),
        ])

        X, y = data_parametrized
        n_input = X.shape[1]
        net.fit(X, y)

        assert net.module_.layer.in_features == n_input
        assert fn_calls == 1

    def test_parameter_name(
        self, net_cls, input_shape_setter_cls, data_parametrized,
    ):
        class MyModule(torch.nn.Module):
            def __init__(self, other_input_dim=22):
                super().__init__()
                self.layer = torch.nn.Linear(other_input_dim, 2)
            def forward(self, X):
                return self.layer(X)

        net = net_cls(MyModule, max_epochs=2, callbacks=[
            input_shape_setter_cls(param_name='other_input_dim'),
        ])

        X, y = data_parametrized
        n_input = X.shape[1]
        net.fit(X, y)

        assert net.module_.layer.in_features == n_input

    def test_module_name(
        self, net_cls, module_cls, input_shape_setter_cls, data_parametrized,
    ):
        class MyNet(net_cls):
            def initialize_module(self):
                kwargs = self.get_params_for('module')
                self.module_ = self.module(**kwargs)

                kwargs = self.get_params_for('module2')
                self.module2_ = self.module(**kwargs)

        net = MyNet(
            module=module_cls,
            max_epochs=2,
            callbacks=[
                input_shape_setter_cls(module_name='module'),
                input_shape_setter_cls(module_name='module2'),
            ],
        )

        X, y = data_parametrized
        n_input = X.shape[1]
        net.fit(X, y)

        assert net.module_.layer.in_features == n_input
        assert net.module2_.layer.in_features == n_input

    def test_no_module_reinit_when_already_correct(
        self, net_cls, module_cls, input_shape_setter_cls, data_fixed,
    ):
        with patch('skorch.classifier.NeuralNetClassifier.initialize_module',
                   side_effect=net_cls.initialize_module, autospec=True):
            net = net_cls(
                module_cls, max_epochs=2, callbacks=[input_shape_setter_cls()],

                # set the input dim to the correct shape beforehand
                module__input_dim=data_fixed[0].shape[-1],
            )

            net.fit(*data_fixed)

            # first initialization due to `initialize()` but not
            # a second one since the input shape is already correct.
            assert net.initialize_module.call_count == 1

    def test_no_module_reinit_partial_fit(
        self, net_cls, module_cls, input_shape_setter_cls, data_fixed,
    ):
        with patch('skorch.classifier.NeuralNetClassifier.initialize_module',
                   side_effect=net_cls.initialize_module, autospec=True):
            net = net_cls(
                module_cls, max_epochs=2, callbacks=[input_shape_setter_cls()],
            )

            net.fit(*data_fixed)
            # first initialization due to `initialize()`, second
            # by setting the input dimension in `on_train_begin`
            assert net.initialize_module.call_count == 2

            net.partial_fit(*data_fixed)
            # no re-initialization when there was no change in
            # input dimension.
            assert net.initialize_module.call_count == 2



"""Benchmark to test time and memory performance of History.

Before #312, the timing would be roughly 5 sec and memory usage would
triple. After #312, the timing would be roughly 2 sec and memory usage
roughly constant.

For the reasons, see #306.

"""

from pprint import pprint
import time

import numpy as np
from sklearn.datasets import make_classification
import torch

from skorch import NeuralNetClassifier
from skorch.callbacks import Callback
from skorch.toy import make_classifier


side_effects = []


class TriggerKeyError(Callback):
    def on_batch_end(self, net, **kwargs):
        try:
            net.history[-1, 'batches', -1, 'foobar']
        except Exception as e:
            pass


class PrintMemory(Callback):
    def on_batch_end(self, net, **kwargs):
        side_effects.append((
            torch.cuda.memory_allocated() / 1e6,
            torch.cuda.memory_cached() / 1e6
        ))


def train():
    X, y = make_classification(1000, 20, n_informative=10, random_state=0)
    X = X.astype(np.float32)
    y = y.astype(np.int64)

    module = make_classifier(input_units=20)

    net = NeuralNetClassifier(
        module,
        max_epochs=10,
        lr=0.1,
        callbacks=[TriggerKeyError(), PrintMemory()],
        device='cuda',
    )

    return net.fit(X, y)


def safe_slice(history, keys):
    # catch errors
    for key in keys:
        try:
            history[key]
        except (KeyError, IndexError):
            pass


def performance_history(history):
    # SUCCESSFUL
    # level 0
    for i in range(len(history)):
        history[i]

    # level 1
    keys = tuple(history[0].keys())
    history[0, keys]
    history[:, keys]
    for key in keys:
        history[0, key]
        history[:, key]

    # level 2
    for i in range(len(history[0, 'batches'])):
        history[0, 'batches', i]
        history[:, 'batches', i]
    history[:, 'batches', :]

    # level 3
    keys = tuple(history[0, 'batches', 0].keys())
    history[0, 'batches', 0, keys]
    history[:, 'batches', 0, keys]
    history[0, 'batches', :, keys]
    history[:, 'batches', :, keys]
    for key in history[0, 'batches', 0]:
        history[0, 'batches', 0, key]
        history[:, 'batches', 0, key]
        history[0, 'batches', :, key]
        history[:, 'batches', :, key]

    # KEY ERRORS
    # level 0
    safe_slice(history, [100000])

    # level 1
    safe_slice(history, [np.s_[0, 'foo'], np.s_[:, 'foo']])

    # level 2
    safe_slice(history, [
        np.s_[0, 'batches', 0],
        np.s_[:, 'batches', 0],
        np.s_[0, 'batches', :],
        np.s_[:, 'batches', :],
    ])

    # level 3
    safe_slice(history, [
        np.s_[0, 'batches', 0, 'foo'],
        np.s_[:, 'batches', 0, 'foo'],
        np.s_[0, 'batches', :, 'foo'],
        np.s_[:, 'batches', :, 'foo'],
        np.s_[0, 'batches', 0, ('foo', 'bar')],
        np.s_[:, 'batches', 0, ('foo', 'bar')],
        np.s_[0, 'batches', :, ('foo', 'bar')],
        np.s_[:, 'batches', :, ('foo', 'bar')],
    ])

if __name__ == '__main__':
    net = train()
    tic = time.time()
    for _ in range(1000):
        performance_history(net.history)
    toc = time.time()
    print("Time for performing 1000 runs: {:.5f} sec.".format(toc - tic))
    assert toc - tic < 10, "accessing history is too slow"

    print("Allocated / cached memory")
    pprint(side_effects)

    mem_start = side_effects[0][0]
    mem_end = side_effects[-1][0]

    print("Memory epoch 1: {:.4f}, last epoch: {:.4f}".format(
        mem_start, mem_end))
    assert np.isclose(mem_start, mem_end, rtol=1/3), "memory use should be similar"



"""Tests for lr_scheduler.py"""
from unittest.mock import Mock

import numpy as np
import pytest
import torch
from sklearn.base import clone
from torch.optim import SGD
from torch.optim.lr_scheduler import CosineAnnealingLR
from torch.optim.lr_scheduler import ExponentialLR
from torch.optim.lr_scheduler import LambdaLR
from torch.optim.lr_scheduler import MultiStepLR
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.lr_scheduler import StepLR
from torch.optim.lr_scheduler import CyclicLR as TorchCyclicLR

from skorch import NeuralNetClassifier
from skorch.callbacks.lr_scheduler import WarmRestartLR, LRScheduler


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
class TestLRCallbacks:

    @pytest.mark.parametrize('policy', [StepLR, 'StepLR'])
    def test_simulate_lrs_epoch_step(self, policy):
        lr_sch = LRScheduler(policy, step_size=2)
        lrs = lr_sch.simulate(6, 1)
        expected = np.array([1.0, 1.0, 0.1, 0.1, 0.01, 0.01])
        assert np.allclose(expected, lrs)

    @pytest.mark.parametrize('policy', [TorchCyclicLR])
    def test_simulate_lrs_batch_step(self, policy):
        lr_sch = LRScheduler(
            policy, base_lr=1, max_lr=5, step_size_up=4, step_every='batch')
        lrs = lr_sch.simulate(11, 1)
        expected = np.array([1, 2, 3, 4, 5, 4, 3, 2, 1, 2, 3])
        assert np.allclose(expected, lrs)

    @pytest.mark.parametrize('policy, instance, kwargs', [
        ('LambdaLR', LambdaLR, {'lr_lambda': (lambda x: 1e-1)}),
        ('StepLR', StepLR, {'step_size': 30}),
        ('MultiStepLR', MultiStepLR, {'milestones': [30, 90]}),
        ('ExponentialLR', ExponentialLR, {'gamma': 0.1}),
        ('ReduceLROnPlateau', ReduceLROnPlateau, {}),
        ('WarmRestartLR', WarmRestartLR, {}),
        ('CosineAnnealingLR', CosineAnnealingLR, {'T_max': 5, 'eta_min': 1e-3}),
        (WarmRestartLR, WarmRestartLR, {}),
    ])
    def test_lr_callback_init_policies(
            self,
            classifier_module,
            classifier_data,
            policy,
            instance,
            kwargs,
    ):
        X, y = classifier_data
        lr_policy = LRScheduler(policy, **kwargs)
        net = NeuralNetClassifier(
            classifier_module, max_epochs=2, callbacks=[lr_policy]
        )
        net.fit(X, y)
        assert any(list(map(
            lambda x: isinstance(
                getattr(x[1], 'lr_scheduler_', None), instance),
            net.callbacks_
        )))

    @pytest.mark.parametrize('policy, kwargs', [
        ('LambdaLR', {'lr_lambda': (lambda x: 1e-1)}),
        ('StepLR', {'step_size': 30}),
        ('MultiStepLR', {'milestones': [30, 90]}),
        ('ExponentialLR', {'gamma': 0.1}),
        ('ReduceLROnPlateau', {}),
        ('WarmRestartLR', {}),
        ('CosineAnnealingLR', {'T_max': 3}),
    ])
    def test_lr_callback_steps_correctly(
            self,
            classifier_module,
            classifier_data,
            policy,
            kwargs,
    ):
        max_epochs = 2
        X, y = classifier_data
        lr_policy = LRScheduler(policy, **kwargs)
        net = NeuralNetClassifier(
            classifier_module(),
            max_epochs=max_epochs,
            batch_size=16,
            callbacks=[lr_policy],
        )
        net.fit(X, y)
        # pylint: disable=protected-access
        assert lr_policy.lr_scheduler_.last_epoch == max_epochs

    @pytest.mark.parametrize('policy, kwargs', [
        (TorchCyclicLR, {'base_lr': 1e-3, 'max_lr': 6e-3, 'step_every': 'batch'}),
    ])
    def test_lr_callback_batch_steps_correctly(
            self,
            classifier_module,
            classifier_data,
            policy,
            kwargs,
    ):
        batch_size = 100
        max_epochs = 2

        X, y = classifier_data
        num_examples = len(X)

        lr_policy = LRScheduler(policy, **kwargs)
        net = NeuralNetClassifier(classifier_module(), max_epochs=max_epochs,
                                  batch_size=batch_size, callbacks=[lr_policy])
        net.fit(X, y)

        total_iterations_per_epoch = num_examples / batch_size
        # 80% of sample used for training by default
        total_training_iterations_per_epoch = 0.8 * total_iterations_per_epoch

        expected = int(total_training_iterations_per_epoch * max_epochs)
        # pylint: disable=protected-access
        assert lr_policy.batch_idx_ == expected

    @pytest.mark.parametrize('policy, kwargs', [
        (TorchCyclicLR, {'base_lr': 1e-3, 'max_lr': 6e-3, 'step_every': 'batch'}),
    ])
    def test_lr_callback_batch_steps_correctly_fallback(
            self,
            classifier_module,
            classifier_data,
            policy,
            kwargs,
    ):
        batch_size = 100
        max_epochs = 2

        X, y = classifier_data
        num_examples = len(X)

        lr_policy = LRScheduler(policy, **kwargs)
        net = NeuralNetClassifier(classifier_module(), max_epochs=max_epochs,
                                  batch_size=batch_size, callbacks=[lr_policy])
        net.fit(X, y)

        # Removes batch count information in the last two epochs
        for i in range(max_epochs):
            del net.history[i]["train_batch_count"]
            del net.history[i]["valid_batch_count"]
        net.partial_fit(X, y)

        total_iterations_per_epoch = num_examples / batch_size

        # batch_counts were removed thus the total iterations of the last
        # epoch is used
        total_iterations_fit_run = total_iterations_per_epoch * max_epochs

        # 80% of sample used for training by default
        total_iterations_partial_fit_run = (
                0.8 * total_iterations_per_epoch * max_epochs)

        # called fit AND partial_fit
        total_iterations = (total_iterations_fit_run +
                            total_iterations_partial_fit_run)
        # Failback to using both valid and training batches counts on
        # second run
        expected = int(total_iterations)
        # pylint: disable=protected-access
        assert lr_policy.batch_idx_ == expected

    def test_lr_scheduler_cloneable(self):
        # reproduces bug #271
        scheduler = LRScheduler(WarmRestartLR, base_lr=123)
        clone(scheduler)  # does not raise

    def test_lr_scheduler_set_params(self, classifier_module, classifier_data):
        scheduler = LRScheduler(
            TorchCyclicLR, base_lr=123, max_lr=999, step_every='batch')
        net = NeuralNetClassifier(
            classifier_module,
            max_epochs=0,
            callbacks=[('scheduler', scheduler)],
        )
        net.set_params(callbacks__scheduler__base_lr=456)
        net.fit(*classifier_data)  # we need to trigger on_train_begin
        assert net.callbacks[0][1].lr_scheduler_.base_lrs[0] == 456

    @pytest.mark.parametrize('policy,kwargs', [
        (StepLR, {'gamma': 0.9, 'step_size': 1})
    ])
    def test_lr_scheduler_record_epoch_step(self,
                                            classifier_module,
                                            classifier_data,
                                            policy,
                                            kwargs):
        epochs = 3
        scheduler = LRScheduler(policy, **kwargs)
        lrs = scheduler.simulate(epochs, initial_lr=123.)
        net = NeuralNetClassifier(
            classifier_module,
            max_epochs=epochs,
            lr=123.,
            callbacks=[('scheduler', scheduler)]
        )
        net.fit(*classifier_data)
        assert np.all(net.history[:, 'event_lr'] == lrs)

    def test_lr_scheduler_record_batch_step(self, classifier_module, classifier_data):
        X, y = classifier_data
        batch_size = 128

        scheduler = LRScheduler(
            TorchCyclicLR,
            base_lr=1,
            max_lr=5,
            step_size_up=4,
            step_every='batch'
        )
        net = NeuralNetClassifier(
            classifier_module,
            max_epochs=1,
            lr=123.,
            batch_size=batch_size,
            callbacks=[('scheduler', scheduler)]
        )
        net.fit(X, y)
        new_lrs = scheduler.simulate(
            net.history[-1, 'train_batch_count'],
            initial_lr=123.,
        )
        assert np.all(net.history[-1, 'batches', :, 'event_lr'] == new_lrs)


class TestReduceLROnPlateau:
    def get_net_with_mock(
            self, classifier_data, classifier_module, monitor='train_loss'):
        """Returns a net with a mocked lr policy that allows to check what
        it's step method was called with.

        """
        X, y = classifier_data
        net = NeuralNetClassifier(
            classifier_module,
            callbacks=[
                ('scheduler', LRScheduler(ReduceLROnPlateau, monitor=monitor)),
            ],
            max_epochs=1,
        ).fit(X, y)

        # mock the policy
        policy = dict(net.callbacks_)['scheduler'].lr_scheduler_
        mock_step = Mock(side_effect=policy.step)
        policy.step = mock_step

        # make sure that mocked policy is set
        scheduler = dict(net.callbacks_)['scheduler']
        # pylint: disable=protected-access
        scheduler._get_scheduler = lambda *args, **kwargs: policy

        net.partial_fit(X, y)
        return net, mock_step

    @pytest.mark.parametrize('monitor', ['train_loss', 'valid_loss', 'epoch'])
    def test_reduce_lr_monitor_with_string(
            self, monitor, classifier_data, classifier_module):
        # step should be called with the 2nd to last value from that
        # history entry
        net, mock_step = self.get_net_with_mock(
            classifier_data, classifier_module, monitor=monitor)
        score = mock_step.call_args_list[0][0][0]
        np.isclose(score, net.history[-2, monitor])

    def test_reduce_lr_monitor_with_callable(
            self, classifier_data, classifier_module):
        # step should always be called with the return value from the
        # callable, 55
        _, mock_step = self.get_net_with_mock(
            classifier_data, classifier_module, monitor=lambda x: 55)
        score = mock_step.call_args_list[0][0][0]
        assert score == 55

    @pytest.mark.parametrize('mode', ['min', 'max'])
    def test_reduce_lr_monitor_passes_monitored_loss(
            self, classifier_data, classifier_module, mode):
        X, y = classifier_data
        net = NeuralNetClassifier(
            classifier_module,
            callbacks=[
                ('scheduler', LRScheduler(
                    ReduceLROnPlateau, monitor='valid_loss', mode=mode)),
            ],
            max_epochs=1,
        )
        net.fit(X, y)

        expected = net.history_[-1, "valid_loss"]
        policy = dict(net.callbacks_)['scheduler'].lr_scheduler_
        assert policy.best == pytest.approx(expected)

    def test_reduce_lr_raise_error_when_key_does_not_exist(
            self, classifier_data, classifier_module):
        X, y = classifier_data
        net = NeuralNetClassifier(
            classifier_module,
            callbacks=[
                ('scheduler', LRScheduler(
                    ReduceLROnPlateau, monitor='bad_key')),
            ],
            max_epochs=1,
        )
        msg = ("'bad_key' was not found in history. A Scoring "
               "callback with name='bad_key' should be placed before the "
               "LRScheduler callback")
        with pytest.raises(ValueError, match=msg):
            net.fit(X, y)


class TestWarmRestartLR():
    def assert_lr_correct(
            self, optimizer, targets, epochs, min_lr, max_lr, base_period,
            period_mult):
        """Test that learning rate was set correctly."""
        targets = [targets] if len(optimizer.param_groups) == 1 else targets
        scheduler = WarmRestartLR(
            optimizer, min_lr, max_lr, base_period, period_mult
        )
        for epoch in range(epochs):
            optimizer.step()  # suppress warning about .step call order
            scheduler.step(epoch)
            for param_group, target in zip(optimizer.param_groups, targets):
                assert param_group['lr'] == pytest.approx(target[epoch])

    def _single_period_targets(self, epochs, min_lr, max_lr, period):
        targets = 1 + np.cos(np.arange(epochs) * np.pi / period)
        targets = min_lr + 0.5 * (max_lr - min_lr) * targets
        return targets.tolist()

    # pylint: disable=missing-docstring
    def _multi_period_targets(
            self, epochs, min_lr, max_lr, base_period, period_mult):
        remaining_epochs = epochs
        current_period = base_period
        targets = list()
        while remaining_epochs > 0:
            period_epochs = min(remaining_epochs, current_period + 1)
            remaining_epochs -= period_epochs
            targets += self._single_period_targets(
                period_epochs, min_lr, max_lr, current_period
            )
            current_period = current_period * period_mult
        return targets

    @pytest.fixture()
    def init_optimizer(self, classifier_module):
        return SGD(classifier_module().parameters(), lr=0.05)

    def test_raise_incompatible_len_on_min_lr_err(self, init_optimizer):
        with pytest.raises(ValueError) as excinfo:
            WarmRestartLR(init_optimizer, min_lr=[1e-1, 1e-2])
        assert 'min_lr' in str(excinfo.value)

    def test_raise_incompatible_len_on_max_lr_err(self, init_optimizer):
        with pytest.raises(ValueError) as excinfo:
            WarmRestartLR(init_optimizer, max_lr=[1e-1, 1e-2])
        assert 'max_lr' in str(excinfo.value)

    def test_single_period(self, init_optimizer):
        optimizer = init_optimizer
        epochs = 3
        min_lr = 5e-5
        max_lr = 5e-2
        base_period = 3
        period_mult = 1
        targets = self._single_period_targets(
            epochs, min_lr, max_lr, base_period)
        self.assert_lr_correct(
            optimizer,
            targets,
            epochs,
            min_lr,
            max_lr,
            base_period,
            period_mult
        )

    def test_multi_period_with_restart(self, init_optimizer):
        optimizer = init_optimizer
        epochs = 9
        min_lr = 5e-5
        max_lr = 5e-2
        base_period = 2
        period_mult = 2
        targets = self._multi_period_targets(
            epochs, min_lr, max_lr, base_period, period_mult
        )
        self.assert_lr_correct(
            optimizer,
            targets,
            epochs,
            min_lr,
            max_lr,
            base_period,
            period_mult
        )

    def test_restarts_with_multiple_groups(self, classifier_module):
        classifier = classifier_module()
        optimizer = SGD(
            [
                {'params': classifier.sequential[0].parameters(), 'lr': 1e-3},
                {'params': classifier.sequential[1].parameters(), 'lr': 1e-2},
                {'params': classifier.sequential[2].parameters(), 'lr': 1e-1},
            ]
        )

        epochs = 9
        min_lr_group = [1e-5, 1e-4, 1e-3]
        max_lr_group = [1e-3, 1e-2, 1e-1]
        base_period = 2
        period_mult = 2
        targets = list()
        for min_lr, max_lr in zip(min_lr_group, max_lr_group):
            targets.append(
                self._multi_period_targets(
                    epochs, min_lr, max_lr, base_period, period_mult
                )
            )
        self.assert_lr_correct(
            optimizer,
            targets,
            epochs,
            min_lr_group,
            max_lr_group,
            base_period,
            period_mult
        )



import os

from skorch.utils import to_device

import torch
from torch.autograd import Variable


class Dictionary:
    def __init__(self):
        self.word2idx = {}
        self.idx2word = []

    def add_word(self, word):
        if word not in self.word2idx:
            self.idx2word.append(word)
            self.word2idx[word] = len(self.idx2word) - 1
        return self.word2idx[word]

    def __len__(self):
        return len(self.idx2word)


class Corpus:
    def __init__(self, path):
        self.dictionary = Dictionary()
        self.train = self.tokenize(os.path.join(path, 'train.txt'))
        self.valid = self.tokenize(os.path.join(path, 'valid.txt'))
        self.test = self.tokenize(os.path.join(path, 'test.txt'))

    def tokenize(self, path):
        """Tokenizes a text file."""
        assert os.path.exists(path)
        # Add words to the dictionary
        with open(path, 'r') as f:
            tokens = 0
            for line in f:
                words = line.split() + ['<eos>']
                tokens += len(words)
                for word in words:
                    self.dictionary.add_word(word)

        # Tokenize file content
        with open(path, 'r') as f:
            ids = torch.LongTensor(tokens)
            token = 0
            for line in f:
                words = line.split() + ['<eos>']
                for word in words:
                    ids[token] = self.dictionary.word2idx[word]
                    token += 1

        return ids


class Loader:
    def __init__(self, source, device='cpu', bptt=10, batch_size=20, evaluation=False):
        self.evaluation = evaluation
        self.bptt = bptt
        self.batch_size = batch_size
        self.device = device
        if isinstance(source.X, Variable):
            data = source.X.data.long()
        else:
            data = torch.LongTensor(source.X)
        self.batches = self.batchify(data, batch_size)

    def batchify(self, data, bsz):
        # Work out how cleanly we can divide the dataset into bsz parts.
        nbatch = data.size(0) // bsz
        # Trim off any extra elements that wouldn't cleanly fit (remainders).
        data = data.narrow(0, 0, nbatch * bsz)
        # Evenly divide the data across the bsz batches.
        data = data.view(bsz, -1).t().contiguous()
        return to_device(data, self.device)

    def get_batch(self, i):
        seq_len = min(self.bptt, len(self.batches) - 1 - i)
        data = Variable(self.batches[i:i+seq_len], volatile=self.evaluation)
        target = Variable(self.batches[i+1:i+1+seq_len].view(-1))
        return data, target

    def __iter__(self):
        for i in range(0, self.batches.size(0) - 1, self.bptt):
            yield self.get_batch(i)



"""Dataset for cells"""
from itertools import zip_longest, product, chain
import random

import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision.transforms.functional import (pad, to_tensor, normalize,
                                               hflip, vflip, crop)
from PIL import Image


def calcuate_bboxes(im_shape, patch_size):
    """Calculate bound boxes based on image shape and size of the bounding box
    given by `patch_size`"""
    h, w = im_shape
    ph, pw = patch_size

    steps_h = chain(range(0, h - ph, ph), [h - ph])
    steps_w = chain(range(0, w - pw, pw), [w - pw])

    return product(steps_h, steps_w)


class PatchedDataset(Dataset):
    """Creates patches of cells.

    Parameters
    ----------
    base_dataset: CellsDataset
        Dataset of cells
    patch_size: tuple of ints (default=(256, 256))
        The size of each patch
    random_flips: bool (default=False)
        If true, patches and masks will be randomly flipped horizontally and
        vertically.
    padding: int (default=16)
        Amount of paddding around each image and mask
    """
    def __init__(self,
                 base_dataset,
                 patch_size=(256, 256),
                 random_flips=False,
                 padding=16):
        super().__init__()
        self.base_dataset = base_dataset
        self.patch_size = patch_size
        self.patch_size_expanded = (patch_size[0] + 2 * padding,
                                    patch_size[1] + 2 * padding)
        self.padding = padding
        self.random_flips = random_flips

        coords = []
        for idx, (_, mask) in enumerate(self.base_dataset):
            w, h = mask.size
            bboxes = calcuate_bboxes((h, w), self.patch_size)
            idx_bboxes = list(zip_longest([], bboxes, fillvalue=idx))
            coords.extend(idx_bboxes)
        self.coords = coords

    def __len__(self):
        return len(self.coords)

    def __getitem__(self, idx):
        img_idx, (i, j) = self.coords[idx]
        cell, mask = self.base_dataset[img_idx]
        h, w = self.patch_size_expanded

        cell = pad(cell, self.padding, padding_mode='reflect')
        mask = pad(mask, self.padding, padding_mode='reflect')

        cell = crop(cell, i, j, h, w)
        mask = crop(mask, i, j, h, w)

        if self.random_flips:
            if random.random() < 0.5:
                cell = hflip(cell)
                mask = hflip(mask)

            if random.random() < 0.5:
                cell = vflip(cell)
                mask = vflip(mask)

        cell = to_tensor(cell)
        mask = torch.as_tensor((np.array(mask) == 255).astype('float32'))

        # mean and std of imagenet
        cell = normalize(cell, [0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

        return cell, mask


class CellsDataset(Dataset):
    """Constructs cell dataset"""
    def __init__(self, sample_dirs):
        super().__init__()
        self.sample_dirs = sample_dirs

    def __len__(self):
        return len(self.sample_dirs)

    def __getitem__(self, idx):
        sample_dir = self.sample_dirs[idx]
        cell_fn = (sample_dir / 'images' / sample_dir.name).with_suffix('.png')
        mask_fn = sample_dir / 'mask.png'

        cell, mask = Image.open(cell_fn).convert('RGB'), Image.open(mask_fn)
        assert cell.size == mask.size
        return cell, mask



"""Contains history class and helper functions."""

import json

from skorch.utils import open_file_like


# pylint: disable=invalid-name
class _none:
    """Special placeholder since ``None`` is a valid value."""


def _not_none(items):
    """Whether the item is a placeholder or contains a placeholder."""
    if not isinstance(items, (tuple, list)):
        items = (items,)
    return all(item is not _none for item in items)


def _getitem_list_list(items, keys, tuple_=False):
    """Ugly but efficient extraction of multiple values from a list of
    items.
    """
    filtered = []
    for item in items:
        row = []
        for key in keys:
            try:
                row.append(item[key])
            except KeyError:
                break
        else:  # no break
            if row:
                filtered.append(tuple(row) if tuple_ else row)
    if items and not filtered:
        return _none
    return filtered


def _getitem_list_tuple(items, keys):
    return _getitem_list_list(items, keys, tuple_=True)


def _getitem_list_str(items, key):
    filtered = []
    for item in items:
        try:
            filtered.append(item[key])
        except KeyError:
            continue
    if items and not filtered:
        return _none
    return filtered


def _getitem_dict_list(item, keys):
    return [item.get(key, _none) for key in keys]


def _getitem_dict_tuple(item, keys):
    return tuple(item.get(key, _none) for key in keys)


def _getitem_dict_str(item, key):
    return item.get(key, _none)


def _get_getitem_method(items, key):
    """Return method to extract values from items.

    For the given type of items and type of keys, find the correct
    method to extract the values. By calling this only once per items,
    we can save a lot of type checking, which can be slow if there are
    a lot of epochs and a lot of items. However, we now make the
    assumption that the type of items doesn't change (we know that the
    key doesn't change). This should always be true, except if
    something really weird happens.

    We are multi-dispatching based on the following possibilities:

    * history[0, 'foo', :10]: get a list of items
    * history[0, 'foo', 0]: get a dict
    * history[0, 'foo', :, 'bar']: key is a str
    * history[0, 'foo', :, ('bar', 'baz')]: key is list/tuple of str

    """
    if isinstance(items, list):
        if isinstance(key, list):
            return _getitem_list_list
        if isinstance(key, tuple):
            return _getitem_list_tuple
        if isinstance(key, str):
            return _getitem_list_str
        raise TypeError("History access with given types not supported")

    if isinstance(items, dict):
        if isinstance(key, list):
            return _getitem_dict_list
        if isinstance(key, tuple):
            return _getitem_dict_tuple
        if isinstance(key, str):
            return _getitem_dict_str
    raise TypeError("History access with given types not supported")


def _unpack_index(i):
    """Unpack index and return exactly four elements.

    If index is more shallow than 4, return None for trailing
    dimensions. If index is deeper than 4, raise a KeyError.

    """
    if len(i) > 4:
        raise KeyError(
            "Tried to index history with {} indices but only "
            "4 indices are possible.".format(len(i)))

    # fill trailing indices with None
    i_e, k_e, i_b, k_b = i + tuple([None] * (4 - len(i)))

    return i_e, k_e, i_b, k_b


class History(list):
    """History contains the information about the training history of
    a :class:`.NeuralNet`, facilitating some of the more common tasks
    that are occur during training.

    When you want to log certain information during training (say, a
    particular score or the norm of the gradients), you should write
    them to the net's history object.

    It is basically a list of dicts for each epoch, that, again,
    contains a list of dicts for each batch. For convenience, it has
    enhanced slicing notation and some methods to write new items.

    To access items from history, you may pass a tuple of up to four
    items:

      1. Slices along the epochs.
      2. Selects columns from history epochs, may be a single one or a
         tuple of column names.
      3. Slices along the batches.
      4. Selects columns from history batchs, may be a single one or a
         tuple of column names.

    You may use a combination of the four items.

    If you select columns that are not present in all epochs/batches,
    only those epochs/batches are chosen that contain said columns. If
    this set is empty, a ``KeyError`` is raised.

    Examples
    --------
    >>> # ACCESSING ITEMS
    >>> # history of a fitted neural net
    >>> history = net.history
    >>> # get current epoch, a dict
    >>> history[-1]
    >>> # get train losses from all epochs, a list of floats
    >>> history[:, 'train_loss']
    >>> # get train and valid losses from all epochs, a list of tuples
    >>> history[:, ('train_loss', 'valid_loss')]
    >>> # get current batches, a list of dicts
    >>> history[-1, 'batches']
    >>> # get latest batch, a dict
    >>> history[-1, 'batches', -1]
    >>> # get train losses from current batch, a list of floats
    >>> history[-1, 'batches', :, 'train_loss']
    >>> # get train and valid losses from current batch, a list of tuples
    >>> history[-1, 'batches', :, ('train_loss', 'valid_loss')]

    >>> # WRITING ITEMS
    >>> # add new epoch row
    >>> history.new_epoch()
    >>> # add an entry to current epoch
    >>> history.record('my-score', 123)
    >>> # add a batch row to the current epoch
    >>> history.new_batch()
    >>> # add an entry to the current batch
    >>> history.record_batch('my-batch-score', 456)
    >>> # overwrite entry of current batch
    >>> history.record_batch('my-batch-score', 789)

    """

    def new_epoch(self):
        """Register a new epoch row."""
        self.append({'batches': []})

    def new_batch(self):
        """Register a new batch row for the current epoch."""
        # pylint: disable=invalid-sequence-index
        self[-1]['batches'].append({})

    def record(self, attr, value):
        """Add a new value to the given column for the current
        epoch.

        """
        msg = "Call new_epoch before recording for the first time."
        if not self:
            raise ValueError(msg)
        self[-1][attr] = value

    def record_batch(self, attr, value):
        """Add a new value to the given column for the current
        batch.

        """
        # pylint: disable=invalid-sequence-index
        self[-1]['batches'][-1][attr] = value

    def to_list(self):
        """Return history object as a list."""
        return list(self)

    @classmethod
    def from_file(cls, f):
        """Load the history of a ``NeuralNet`` from a json file.

        Parameters
        ----------
        f : file-like object or str

        """

        with open_file_like(f, 'r') as fp:
            return cls(json.load(fp))

    def to_file(self, f):
        """Saves the history as a json file. In order to use this feature,
        the history must only contain JSON encodable Python data structures.
        Numpy and PyTorch types should not be in the history.

        Parameters
        ----------
        f : file-like object or str

        """
        with open_file_like(f, 'w') as fp:
            json.dump(self.to_list(), fp)

    def __getitem__(self, i):
        # This implementation resolves indexing backwards,
        # i.e. starting from the batches, then progressing to the
        # epochs.
        if isinstance(i, (int, slice)):
            i = (i,)

        # i_e: index epoch, k_e: key epoch
        # i_b: index batch, k_b: key batch
        i_e, k_e, i_b, k_b = _unpack_index(i)
        keyerror_msg = "Key {!r} was not found in history."

        if i_b is not None and k_e != 'batches':
            raise KeyError("History indexing beyond the 2nd level is "
                           "only possible if key 'batches' is used, "
                           "found key {!r}.".format(k_e))

        items = self.to_list()

        # extract the epochs
        # handles: history[i_e]
        if i_e is not None:
            items = items[i_e]
            if isinstance(i_e, int):
                items = [items]

        # extract indices of batches
        # handles: history[..., k_e, i_b]
        if i_b is not None:
            items = [row[k_e][i_b] for row in items]

        # extract keys of epochs or batches
        # handles: history[..., k_e]
        # handles: history[..., ..., ..., k_b]
        if k_e is not None and (i_b is None or k_b is not None):
            key = k_e if k_b is None else k_b

            if items:
                extract = _get_getitem_method(items[0], key)
                items = [extract(item, key) for item in items]

                # filter out epochs with missing keys
                items = list(filter(_not_none, items))

            if not items and not (k_e == 'batches' and i_b is None):
                # none of the epochs matched
                raise KeyError(keyerror_msg.format(key))

            if (
                    isinstance(i_b, slice)
                    and k_b is not None
                    and not any(batches for batches in items)
            ):
                # none of the batches matched
                raise KeyError(keyerror_msg.format(key))

        if isinstance(i_e, int):
            items, = items

        return items



""" Post-process regularization steps such as gradient normalizing. """

from torch.nn.utils import clip_grad_norm_

from skorch.callbacks import Callback


__all__ = ['GradientNormClipping']


class GradientNormClipping(Callback):
    """Clips gradient norm of a module's parameters.

    The norm is computed over all gradients together, as if they were
    concatenated into a single vector. Gradients are modified
    in-place.

    See :func:`torch.nn.utils.clip_grad_norm_` for more information.

    Parameters
    ----------
    gradient_clip_value : float (default=None)
      If not None, clip the norm of all model parameter gradients to this
      value. The type of the norm is determined by the
      ``gradient_clip_norm_type`` parameter and defaults to L2.

    gradient_clip_norm_type : float (default=2)
      Norm to use when gradient clipping is active. The default is
      to use L2-norm. Can be 'inf' for infinity norm.

    """
    def __init__(
            self,
            gradient_clip_value=None,
            gradient_clip_norm_type=2,
    ):
        self.gradient_clip_value = gradient_clip_value
        self.gradient_clip_norm_type = gradient_clip_norm_type

    def on_grad_computed(self, _, named_parameters, **kwargs):
        if self.gradient_clip_value is None:
            return

        clip_grad_norm_(
            (p for _, p in named_parameters),
            max_norm=self.gradient_clip_value,
            norm_type=self.gradient_clip_norm_type,
        )
